# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2019-2024 Carl Drougge                                     #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from accelerator.dataset import Dataset
from accelerator.build import JobError
from accelerator.compat import monotonic

from collections import OrderedDict
from datetime import date, datetime, timedelta
from sys import exit
import os
import sys

description = '''
Run the tests. Needs at least 3 slices to work.
'''

def main(urd):
	assert urd.info.slices >= 3, "The tests don't work with less than 3 slices (you have %d)." % (urd.info.slices,)

	from accelerator.shell import cfg
	# sys.argv[0] is absolute for unqualified commands ("ax"), but exactly what
	# the user wrote otherwise ("./venv/bin/ax"). This makes either absolute.
	command_prefix = [os.path.join(cfg.user_cwd, sys.argv[0]), '--config', cfg.config_filename]

	print()
	print("Testing urd.build and job.load")
	want = ({'foo': 'foo', 'a': 'a'}, {'foo': '', 'b': ''}, {'foo': '', 'c': ''})
	job = urd.build("test_build_kws")
	assert job.load() == want
	bad = None
	try:
		urd.build("test_build_kws", options=dict(foo='bar'), foo='baz')
		bad = 'Allowed ambiguous keyword "foo"'
	except Exception:
		pass
	assert not bad, bad
	want[0]['foo'] = 'bar'
	want[0]['a'] = 'A'
	job = urd.build("test_build_kws", options=dict(foo='bar'), a='A')
	assert job.load() == want
	assert urd.build("test_build_kws", options=dict(foo='bar'), a='A', b=None, c=None) == job
	want[2]['c'] = job
	job = urd.build("test_build_kws", options=dict(foo='bar', a='override this from kw'), a='A', c=job)
	assert job.load() == want
	want[0]['foo'] = 'foo'
	want[2]['c'] = job
	job = urd.build("test_build_kws", a='A', b=None, c=job, datasets=dict(b='overridden'))
	assert job.load() == want

	print()
	print("Testing urd.begin/end/truncate/get/peek/latest/first/since")
	urd.build('test_urd', command_prefix=command_prefix)
	urd.truncate("tests.urd", 0)
	assert not urd.peek_latest("tests.urd").joblist
	urd.begin("tests.urd", 1, caption="first")
	urd.build("test_build_kws")
	fin = urd.finish("tests.urd")
	assert fin == {'new': True, 'changed': False, 'is_ghost': False}, fin
	urd.begin("tests.urd", 1)
	job = urd.build("test_build_kws")
	fin = urd.finish("tests.urd", caption="first")
	assert fin == {'new': False, 'changed': False, 'is_ghost': False}, fin
	urd.begin("tests.urd", 1) # will be overridden to 2 in finish
	jl = urd.latest("tests.urd").joblist
	assert jl == [job], '%r != [%r]' % (jl, job,)
	urd.build("test_build_kws", options=dict(foo='bar', a='A'))
	urd.finish("tests.urd", 2, caption="second")
	u = urd.peek_latest("tests.urd")
	assert u.caption == "second"
	dep0 = list(u.deps.values())[0]
	assert dep0.caption == "first", dep0.caption
	assert dep0.joblist == jl, '%r != %r' % (dep0.joblist, jl,)
	assert urd.since("tests.urd", 0) == ['1', '2']
	urd.truncate("tests.urd", 2)
	assert urd.since("tests.urd", 0) == ['1']
	urd.truncate("tests.urd", 0)
	assert urd.since("tests.urd", 0) == []
	ordered_ts = [1, 2, 1000000000, '1978-01-01', '1978-01-01+0', '1978-01-01+2', '1978-01-01 00:00', '1978-01-01T00:00+42', '2017-06-27', '2017-06-27T17:00:00', '2017-06-27 17:00:00+42']
	for ts in ordered_ts:
		urd.begin("tests.urd")
		if ts == 1000000000:
			urd.get("tests.urd", '1')
		urd.build("test_build_kws")
		urd.finish("tests.urd", ts)
	urd.begin("tests.urd")
	urd.build("test_build_kws")
	urd.finish("tests.urd", ('2019-12', 3))
	ordered_ts.append('2019-12+3')
	ordered_ts = [str(v).replace(' ', 'T') for v in ordered_ts]
	assert urd.since("tests.urd", 0) == ordered_ts
	assert urd.since("tests.urd", '1978-01-01') == ordered_ts[4:]
	assert urd.peek_first("tests.urd").timestamp == '1'
	assert not urd.peek("tests.urd", 2).deps
	dep_jl = list(urd.peek("tests.urd", 1000000000).deps.values())[0].joblist
	assert dep_jl == [job]
	assert urd.peek("tests.urd", ('2017-06-27 17:00:00', 42)).timestamp == '2017-06-27T17:00:00+42'
	while ordered_ts:
		urd.truncate("tests.urd", ordered_ts.pop())
		assert urd.since("tests.urd", 0) == ordered_ts, ordered_ts
	want = [date.today() - timedelta(10), datetime.utcnow()]
	for ts in want:
		urd.begin("tests.urd", ts)
		urd.build("test_build_kws")
		urd.finish("tests.urd")
	assert urd.since("tests.urd", 0) == [str(ts).replace(' ', 'T') for ts in want]
	urd.truncate("tests.urd", 0)

	for how in ("exiting", "dying",):
		print()
		print("Verifying that an analysis process %s kills the job" % (how,))
		time_before = monotonic()
		try:
			job = urd.build("test_analysis_died", how=how)
			print("test_analysis_died completed successfully (%s), that shouldn't happen" % (job,))
			exit(1)
		except JobError:
			time_after = monotonic()
		time_to_die = time_after - time_before
		if time_to_die > 13:
			print("test_analysis_died took %d seconds to die, it should be faster" % (time_to_die,))
			exit(1)
		elif time_to_die > 2:
			print("test_analysis_died took %d seconds to die, so death detection is slow, but works" % (time_to_die,))
		else:
			print("test_analysis_died took %.1f seconds to die, so death detection works" % (time_to_die,))

	print()
	print("Testing dataset creation, export, import")
	source = urd.build("test_datasetwriter")
	urd.build("test_datasetwriter_verify", source=source)
	source = urd.build("test_datasetwriter_copy", source=source)
	urd.build("test_datasetwriter_verify", source=source)
	urd.build("test_datasetwriter_parent")
	urd.build("test_datasetwriter_missing_slices")
	urd.build("test_datasetwriter_default")
	urd.build("test_datasetwriter_parsed")
	urd.build("test_dataset_in_prepare")
	ds = Dataset(source, "passed")
	csvname = "out.csv.gz"
	csvname_uncompressed = "out.csv"
	csv = urd.build("csvexport", filename=csvname, separator="\t", source=ds)
	csv_uncompressed = urd.build("csvexport", filename=csvname_uncompressed, separator="\t", source=ds)
	csv_quoted = urd.build("csvexport", filename=csvname, quote_fields='"', source=ds)
	urd.build("csvexport", filename='slice%d.csv', sliced=True, source=ds) # unused
	reimp_csv = urd.build("csvimport", filename=csv.filename(csvname), separator="\t")
	reimp_csv_uncompressed = urd.build("csvimport", filename=csv_uncompressed.filename(csvname_uncompressed), separator="\t")
	reimp_csv_quoted = urd.build("csvimport", filename=csv_quoted.filename(csvname), quotes=True)
	urd.build("test_compare_datasets", a=reimp_csv, b=reimp_csv_uncompressed)
	urd.build("test_compare_datasets", a=reimp_csv, b=reimp_csv_quoted)

	print()
	print("Testing job related functions")
	urd.build("test_subjobs_type", typed=ds, untyped=reimp_csv)
	urd.build("test_subjobs_nesting")
	urd.build('test_finish_early')
	urd.build('test_register_file')

	print()
	print("Testing datasets more")
	dsnamejob = urd.build("test_dataset_names")
	# make sure .datasets works with these names (only possible after job finishes)
	assert [ds.name for ds in dsnamejob.datasets] == dsnamejob.load()
	urd.build("test_dataset_column_names")
	mergejob = urd.build("test_dataset_merge")
	urd.build('test_dataset_fanout')
	urd.build("test_dataset_filter_columns")
	urd.build("test_dataset_empty_colname")
	urd.build("test_dataset_nan")
	urd.build('test_dataset_parsing_writer')
	urd.build('test_dataset_overwrite')
	urd.build('test_dataset_rename_columns')
	urd.build('test_dataset_concat')

	print()
	print("Testing order preservation in dicts in options")
	for key_order in (['a', 'c', 'b', 'z', 'd', 'inner'], ['foo', 'bar', 'inner', 'aaa']):
		d = OrderedDict((k, 0) for k in key_order)
		d['inner'] = OrderedDict((k, 1) for k in key_order)
		urd.build("test_options_dict_order", dict=d, key_order=key_order)
	print("And that changing order rebuilds when expected")
	chars4rebuild_test = 'bfhnaeolkgjdmci' # same order as the default in the method
	rb_noopt  = urd.build('test_options_rebuild')
	rb_same   = urd.build('test_options_rebuild', dict=OrderedDict([(c, c) for c in chars4rebuild_test]))
	rb_unord  = urd.build('test_options_rebuild', dict={c: c for c in chars4rebuild_test}) # not ordered, so will be sorted
	rb_sorted = urd.build('test_options_rebuild', dict=OrderedDict([(c, c) for c in sorted(chars4rebuild_test)]))
	rb_other  = urd.build('test_options_rebuild', dict=OrderedDict([(c, c) for c in reversed(chars4rebuild_test)]))
	assert rb_noopt == rb_same  , 'OrderedDict rebuilt job equivalent to default'
	assert rb_unord == rb_sorted, 'OrderedDict rebuilt equivalent non-ordered job'
	assert rb_noopt != rb_unord , 'OrderedDict did not rebuild job with different order'
	assert rb_noopt != rb_other , 'OrderedDict did not rebuild job with different order'

	print()
	print("Testing ['lists'] of datasets and jobs")
	urd.build(
		"test_arg_lists",
		dslist=[
			'',
			mergejob.dataset('a0'),
			mergejob.dataset('a1'),
			mergejob.dataset('b0'),
			mergejob.dataset('b1'),
			mergejob.dataset('c0'),
			mergejob.dataset('c1'),
			None,
		],
		joblist=[None, source, ''],
	)

	print()
	print("Testing csvimport with more difficult files")
	urd.build("test_csvimport_corner_cases")
	urd.build("test_csvimport_separators")
	urd.build("test_csvimport_slicing")

	print()
	print("Testing csvexport with all column types, strange separators, ...")
	urd.build("test_csvexport_naming")
	urd.build("test_csvexport_all_coltypes")
	urd.build("test_csvexport_separators")
	urd.build("test_csvexport_chains")
	urd.build("test_csvexport_quoting")

	print()
	print("Testing dataset typing")
	try:
		# Test if numeric_comma is broken (presumably because no suitable locale
		# was found, since there are not actually any commas in the source dataset.)
		urd.build("dataset_type", source=source, numeric_comma=True, column2type=dict(b="float64"), defaults=dict(b="0"))
		comma_broken = False
	except JobError as e:
		comma_broken = True
		urd.warn()
		urd.warn('SKIPPED NUMERIC COMMA TESTS')
		urd.warn('Follow the instructions in this error to enable numeric comma:')
		urd.warn()
		urd.warn(e.format_msg())
	urd.build("test_dataset_type_None")
	urd.build("test_dataset_type_corner_cases", numeric_comma=not comma_broken)
	urd.build("test_dataset_type_minmax")

	print()
	print("Testing dataset chaining, filtering, callbacks and rechaining")
	selfchain = urd.build("test_selfchain", previous=source)
	urd.build("test_rechain", jobs=dict(selfchain=selfchain))
	urd.build("test_dataset_callbacks")
	urd.build("test_dataset_range")

	print()
	print("Testing dataset sorting and rehashing (with subjobs again)")
	urd.build("test_sorting")
	urd.build("test_sort_stability")
	urd.build("test_sort_chaining")
	urd.build("test_sort_trigger")
	urd.build("test_hashpart")
	urd.build("test_dataset_type_hashing")
	urd.build("test_dataset_type_chaining")

	print()
	print("Test hashlabels")
	urd.build("test_hashlabel")

	print()
	print("Test dataset roundrobin iteration and slicing")
	urd.build("test_dataset_roundrobin")
	urd.build("test_dataset_slice")
	urd.build("test_dataset_unroundrobin")
	urd.build("test_dataset_unroundrobin_trigger")
	urd.build("test_number")
	urd.build("test_nan")

	print()
	print("Test dataset_checksum")
	urd.build("test_dataset_checksum")

	print()
	print("Test csvimport_zip")
	urd.build("test_csvimport_zip")

	print()
	print("Test output handling")
	urd.build("test_output")
	urd.build("test_output_on_error")

	print()
	print("Test special types in options")
	urd.build("test_datetime")
	urd.build("test_path")

	print()
	print("Test various utility functions")
	urd.build("test_optionenum")
	urd.build("test_json")
	urd.build("test_job_save")
	bgsave = urd.build("test_job_save_background")
	assert len(bgsave.files()) == 10 + 4 * urd.info.slices
	urd.build("test_jobwithfile")
	urd.build("test_jobchain")

	print()
	print("Testing that the status stack is correct in exceptions from job building")
	urd.build("test_status_in_exceptions")

	print()
	print("Test shell commands")
	urd.truncate("tests.urd", 0)
	# These have to be rebuilt every time, or the resolving might give other jobs.
	urd.begin("tests.urd", 1)
	a = urd.build('test_shell_data', force_build=True)
	b = urd.build('test_shell_data', force_build=True)
	c = urd.build('test_shell_data', datasets={'previous': a})
	urd.finish("tests.urd")
	urd.begin("tests.urd", "2021-09-27T03:14")
	d = urd.build('test_shell_data', datasets={'previous': c, 'parent': a + '/j'}, jobs={'previous': b, 'extra': c})
	urd.finish("tests.urd")
	urd.begin("tests.urd", "2021-09-27T03:14+1")
	e = urd.build('test_shell_data', jobs={'previous': d, 'extra': a})
	urd.finish("tests.urd")
	urd.build('test_shell_commands', command_prefix=command_prefix)
	# ~ finds earlier jobs with that method, ^ follows jobs.previous falling back to datasets.previous.
	want = {
		'test_shell_data': e, # just the plain method -> job resolution.
		c + '~~': a, # not using .previous, just going back jobs
		'test_shell_data~3': b, # numbered tildes
		'test_shell_data~2^': a, # ~~ goes to c, ^ follows .previous to a.
		d + '^': b, # prefers jobs.previous to .datasets.previous
		d + '.parent': a, # will be different as a ds
		':tests.urd:': e,
		':tests.urd/2021-09-27T03:14:': d,
		':tests.urd/1:1': b, # 1 is the second entry
		':tests.urd/1:-3': a, # third entry from the end
		':tests.urd:^': d,
		':tests.urd/2021-09-27T03:14+1^^:0': a, # ^ in :: goes to earlier entries
		':tests.urd/1~:': d, # ~ in :: goes to later entries
		':tests.urd/2021-09-27T03:14:.extra': c,
		':tests.urd/2021-09-27T03:14+1:.extra': a,
		':tests.urd/2021-09-27T03:14+1:.jobs.previous': d,
		':tests.urd/2021-09-27T03:14+1:.jobs.previous.extra': c,
		':tests.urd/2021-09-27T03:14+1:.jobs.previous.jobs.extra.datasets.previous': a,
		':tests.urd/2021-09-27T03:14+1:.jobs.previous.parent': a, # will be different as a ds
		':tests.urd/2021-09-27T03:14+1:.jobs.previous~.datasets.previous': a,
	}
	urd.build('test_shell_job', command_prefix=command_prefix, want=want)
	# most of the old specs give the same results
	want = {spec: job + '/default' for spec, job in want.items()}
	want.update({
		d + '.parent': a + '/j', # this changes as a ds
		':tests.urd/2021-09-27T03:14+1:.jobs.previous.parent': a + '/j', # and this too
		# below here are new things, not overrides.
		d + '/j^': a + '/j', # .parent
		d + '/j~': b + '/j', # .previous
		'test_shell_data~/j^': a + '/j', # both job and ds movement
		e + '/j~^': a + '/j', # .previous.parent
		# some urdlist ones with datasets on
		':tests.urd:/j': e + '/j',
		':tests.urd/1:1/j': b + '/j',
		':tests.urd:^/j': d + '/j',
		':tests.urd/2021-09-27T03:14:/j': d + '/j',
		# finally one with : in the list and / in the ds name
		':tests.urd/2021-09-27T03:14+1:0/name/with/slash': e + '/name/with/slash',
	})
	urd.build('test_shell_ds', command_prefix=command_prefix, want=want)
	urd.truncate("tests.urd", 0)
	urd.build('test_shell_grep', command_prefix=command_prefix)
	urd.build('test_shell_config', command_prefix=command_prefix)

	print()
	print("Test board")
	urd.build('test_board_metadata', command_prefix=command_prefix)

	summary = urd.build("test_summary", joblist=urd.joblist_all)
	summary.link_result('summary.html')



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2023-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

description = r'''
Test the pathlib types in options.
'''

from pathlib import Path, PurePath, PosixPath, PurePosixPath

from accelerator import subjobs

options = dict(
	inner=False,
	plainPath=Path,
	plainPurePath=PurePath,
	plainPosixPath=PosixPath,
	plainPurePosixPath=PurePosixPath,
	valuedPath=Path('/valued/path'), # Actually a PosixPath
	valuedPurePath=PurePath('valued/pure/path'),
	valuedPosixPath=PosixPath('/valued/posix/path'), # Actually a PurePosixPath
	valuedPurePosixPath=PurePosixPath('valued/pure/posix/path'),
)

def synthesis():
	if options.inner:
		parent = lambda p: p.parent if p else None
		return (
			parent(options.plainPath),
			parent(options.plainPurePath),
			parent(options.plainPosixPath),
			parent(options.plainPurePosixPath),
			parent(options.valuedPath),
			parent(options.valuedPurePath),
			parent(options.valuedPosixPath),
			parent(options.valuedPurePosixPath),
		)

	job = subjobs.build(
		'test_path',
		inner=True,
	)
	assert job.load() == (
		None, None,
		None, None,
		Path('/valued'), PurePath('valued/pure'),
		PosixPath('/valued/posix'), PurePosixPath('valued/pure/posix'),
	)

	job = subjobs.build(
		'test_path',
		inner=True,
		plainPath=Path('hello/there'),
		plainPurePath=Path('over/here'),
		plainPosixPath=PosixPath('/some/where'),
		plainPurePosixPath=PosixPath('there'),
		valuedPath=None,
		valuedPurePath=None,
		valuedPosixPath=None,
		valuedPurePosixPath=None,
	)
	assert job.load() == (
		Path('hello'), PurePath('over'),
		PosixPath('/some'), PurePosixPath('.'),
		None, None,
		None, None,
	)

	# Same again, but with the options as strings.
	str_job = subjobs.build(
		'test_path',
		inner=True,
		plainPath='hello/there',
		plainPurePath='over/here',
		plainPosixPath='/some/where',
		plainPurePosixPath='there',
		valuedPath=None,
		valuedPurePath=None,
		valuedPosixPath=None,
		valuedPurePosixPath=None,
	)
	assert job == str_job



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2024 Carl Drougge                       #
# Modifications copyright (c) 2019-2020 Anders Berkeman                    #
# Modifications copyright (c) 2023-2024 Pablo Correa Gómez                 #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

import os
import re

from collections import namedtuple, OrderedDict
from functools import wraps
from pathlib import Path

from accelerator.compat import unicode, PY2, PY3, open, iteritems, FileNotFoundError
from accelerator.error import NoSuchJobError, NoSuchWorkdirError, NoSuchDatasetError, AcceleratorError


# WORKDIRS should live in the Automata class, but only for callers
# (methods read it too, though hopefully only through the functions in this module)

WORKDIRS = {}


def dirnamematcher(name):
	return re.compile(re.escape(name) + r'-[0-9]+$').match


def _assert_is_normrelpath(path, dirtype):
	norm = os.path.normpath(path)
	if (norm != path and norm + '/' != path) or norm.startswith('/'):
		raise AcceleratorError('%r is not a normalised relative path' % (path,))
	if norm == '..' or norm.startswith('../'):
		raise AcceleratorError('%r is above the %s dir' % (path, dirtype))


def _cachedprop(meth):
	@property
	@wraps(meth)
	def wrapper(self):
		if meth.__name__ not in self._cache:
			self._cache[meth.__name__] = meth(self)
		return self._cache[meth.__name__]
	return wrapper

_cache = {}
_nodefault = object()

class Job(unicode):
	"""
	A string that is a jobid, but also has some extra properties:
	.method The job method (can be the "name" when from build or urd).
	.number The job number as an int.
	.workdir The workdir name (the part before -number in the jobid)
	.path The filesystem directory where the job is stored.
	.params setup.json from this job.
	.post post.json from this job.
	.datasets list of Datasets in this job.
	And some functions:
	.withfile a JobWithFile with this job.
	.filename to join .path with a filename.
	.load to load a pickle.
	.json_load to load a json file.
	.open to open a file (like standard open)
	.dataset to get a named Dataset.
	.output to get what the job printed.
	.link_result to put a link in result_directory that points to a file in this job.
	.link_to_here to expose a subjob result in its parent.

	Decays to a (unicode) string when pickled.
	"""

	__slots__ = ('workdir', 'number', '_cache')

	def __new__(cls, jobid, method=None):
		k = (jobid, method)
		if k in _cache:
			return _cache[k]
		obj = unicode.__new__(cls, jobid)
		try:
			obj.workdir, tmp = jobid.rsplit('-', 1)
			obj.number = int(tmp)
		except ValueError:
			raise NoSuchJobError('Not a valid jobid: "%s".' % (jobid,))
		obj._cache = {}
		if method:
			obj._cache['method'] = method
		_cache[k] = obj
		return obj

	@classmethod
	def _create(cls, name, number):
		return Job('%s-%d' % (name, number,))

	@_cachedprop
	def method(self):
		return self.params.method

	@_cachedprop
	def input_directory(self):
		return self.params.get('input_directory', None)

	@_cachedprop
	def is_build(self):
		return self.params.get('is_build', False)

	@_cachedprop
	def parent(self):
		parent = self.params.get('parent')
		return Job(parent) if parent else NoJob

	@_cachedprop
	def build_job(self):
		candidate = self.parent
		while candidate and not candidate.is_build:
			candidate = candidate.parent
		return candidate

	@property
	def path(self):
		if self.workdir not in WORKDIRS:
			raise NoSuchWorkdirError('Not a valid workdir: "%s"' % (self.workdir,))
		return os.path.join(WORKDIRS[self.workdir], self)

	def filename(self, filename, sliceno=None):
		if isinstance(filename, Path):
			filename = str(filename)
		if sliceno is not None:
			filename = '%s.%d' % (filename, sliceno,)
		return os.path.join(self.path, filename)

	def open(self, filename, mode='r', sliceno=None, encoding=None, errors=None):
		assert 'r' in mode, "Don't write to other jobs"
		if 'b' not in mode and encoding is None:
			encoding = 'utf-8'
		return open(self.filename(filename, sliceno), mode, encoding=encoding, errors=errors)

	def files(self, pattern='*'):
		from fnmatch import filter
		try:
			return set(filter(self.post.files, pattern))
		except FileNotFoundError:
			return set()

	def withfile(self, filename, sliced=False, extra=None):
		return JobWithFile(self, filename, sliced, extra)

	@property # usually cached
	def params(self):
		if 'params' in self._cache:
			return self._cache['params']
		from accelerator.extras import job_params
		res = job_params(self)
		if 'endtime' in res:
			# only cache it if the job is finished, so that board does
			# not get stuck with an incomplete job.
			self._cache['params'] = res
		return res

	@_cachedprop
	def version(self):
		# this is self.params.version, but without fully loading the params
		# (unless already loaded).
		if 'params' in self._cache:
			return self._cache['params'].version
		from accelerator.setupfile import load_setup
		return load_setup(self).version

	@_cachedprop
	def post(self):
		from accelerator.extras import job_post
		return job_post(self)

	def load(self, filename='result.pickle', sliceno=None, encoding='bytes', default=_nodefault):
		"""blob.load from this job"""
		from accelerator.extras import pickle_load
		try:
			return pickle_load(self.filename(filename, sliceno), encoding=encoding)
		except FileNotFoundError:
			if default is _nodefault:
				raise
			return default

	def json_load(self, filename='result.json', sliceno=None, unicode_as_utf8bytes=PY2, default=_nodefault):
		from accelerator.extras import json_load
		try:
			return json_load(self.filename(filename, sliceno), unicode_as_utf8bytes=unicode_as_utf8bytes)
		except FileNotFoundError:
			if default is _nodefault:
				raise
			return default

	def dataset(self, name='default'):
		from accelerator.dataset import Dataset
		return Dataset(self, name)

	@_cachedprop
	def datasets(self):
		from accelerator.dataset import job_datasets
		return job_datasets(self)

	def output(self, what=None):
		if what == 'parts':
			as_parts = True
			what = None
		else:
			as_parts = False
		if isinstance(what, int):
			fns = [what]
		else:
			assert what in (None, 'prepare', 'analysis', 'synthesis'), 'Unknown output %r' % (what,)
			if what in (None, 'analysis'):
				fns = list(range(self.params.slices))
				if what is None:
					fns = ['prepare'] + fns + ['synthesis']
			else:
				fns = [what]
		res = OrderedDict()
		for k in fns:
			fn = self.filename('OUTPUT/' + str(k))
			if os.path.exists(fn):
				with open(fn, 'rt', encoding='utf-8', errors='backslashreplace') as fh:
					res[k] = fh.read()
		if as_parts:
			return res
		else:
			return ''.join(res.values())

	def link_result(self, filename='result.pickle', linkname=None):
		"""Put a symlink to filename in result_directory
		Only use this in a build script."""
		from accelerator.g import running
		assert running == 'build', "Only link_result from a build script"
		from accelerator.shell import cfg
		if isinstance(filename, Path):
			filename = str(filename)
		if isinstance(linkname, Path):
			linkname = str(linkname)
		_assert_is_normrelpath(filename, 'job')
		if linkname is None:
			linkname = os.path.basename(filename.rstrip('/'))
		_assert_is_normrelpath(linkname, 'result')
		if linkname.endswith('/'):
			if filename.endswith('/'):
				linkname = linkname.rstrip('/')
			else:
				linkname += os.path.basename(filename)
		source_fn = os.path.join(self.path, filename)
		assert os.path.exists(source_fn), "Filename \"%s\" does not exist in jobdir \"%s\"!" % (filename, self.path)
		result_directory = cfg['result_directory']
		dest_fn = result_directory
		for part in linkname.split('/'):
			if not os.path.exists(dest_fn):
				os.mkdir(dest_fn)
			elif dest_fn != result_directory and os.path.islink(dest_fn):
				raise AcceleratorError("Refusing to create link %r: %r is a symlink" % (linkname, dest_fn))
			dest_fn = os.path.join(dest_fn, part)
		try:
			os.remove(dest_fn + '_')
		except OSError:
			pass
		os.symlink(source_fn, dest_fn + '_')
		os.rename(dest_fn + '_', dest_fn)

	def link_to_here(self, filename='result.pickle'):
		from accelerator.g import job
		src = self.filename(filename)
		assert os.path.exists(src)
		dst = job.filename(filename)
		os.symlink(src, dst)
		job.register_file(dst)

	def chain(self, length=-1, reverse=False, stop_job=None):
		"""Like Dataset.chain but for jobs."""
		if isinstance(stop_job, dict):
			assert len(stop_job) == 1, "Only pass a single stop_job={job: name}"
			stop_job, stop_name = next(iteritems(stop_job))
			if stop_job:
				stop_job = Job(stop_job).params.jobs.get(stop_name)
		chain = []
		current = self
		while length != len(chain) and current and current != stop_job:
			chain.append(current)
			current = current.params.jobs.get('previous')
		if not reverse:
			chain.reverse()
		return chain

	# Look like a string after pickling
	def __reduce__(self):
		return unicode, (unicode(self),)


class CurrentJob(Job):
	"""The currently running job (as passed to the method),
	with extra functions for writing data."""

	__slots__ = ('input_directory',)

	def __new__(cls, jobid, params):
		obj = Job.__new__(cls, jobid, params.method)
		obj._cache['params'] = params
		obj.input_directory = params.input_directory
		return obj

	def finish_early(self, result=None):
		"""Finish job (successfully) without running later stages"""
		from accelerator.launch import _FinishJob
		raise _FinishJob(result)

	def save(self, obj, filename='result.pickle', sliceno=None, temp=None, background=False):
		from accelerator.extras import pickle_save
		return pickle_save(obj, filename, sliceno, temp=temp, background=background)

	def json_save(self, obj, filename='result.json', sliceno=None, sort_keys=True, temp=None, background=False):
		from accelerator.extras import json_save
		return json_save(obj, filename, sliceno, sort_keys=sort_keys, temp=temp, background=background)

	def datasetwriter(self, columns={}, filename=None, hashlabel=None, hashlabel_override=False, caption=None, previous=None, name='default', parent=None, meta_only=False, for_single_slice=None, copy_mode=False, allow_missing_slices=False):
		from accelerator.dataset import DatasetWriter
		return DatasetWriter(columns=columns, filename=filename, hashlabel=hashlabel, hashlabel_override=hashlabel_override, caption=caption, previous=previous, name=name, parent=parent, meta_only=meta_only, for_single_slice=for_single_slice, copy_mode=copy_mode, allow_missing_slices=allow_missing_slices)

	def open(self, filename, mode='r', sliceno=None, encoding=None, errors=None, temp=None):
		"""Mostly like standard open with sliceno and temp,
		but you must use it as context manager
		with job.open(...) as fh:
		and the file will have a temp name until the with block ends.
		"""
		if 'r' in mode:
			return Job.open(self, filename, mode, sliceno, encoding, errors)
		if 'b' not in mode and encoding is None:
			encoding = 'utf-8'
		if PY3 and 'x' not in mode:
			mode = mode.replace('w', 'x')
		def _open(fn, _mode):
			# ignore the passed mode, use the one we have
			return open(fn, mode, encoding=encoding, errors=errors)
		from accelerator.extras import FileWriteMove
		fwm = FileWriteMove(self.filename(filename, sliceno), temp=temp)
		fwm._open = _open
		return fwm

	def register_file(self, filename):
		"""Record a file produced by this job. Normally you would use
		job.open to have this happen automatically, but if the file was
		produced in a way where that is not practical you can use this
		to register it."""
		filename = self.filename(filename)
		assert os.path.exists(filename)
		from accelerator.extras import saved_files
		saved_files[filename] = 0

	def register_files(self, pattern='**/*' if PY3 else '*'):
		"""Bulk register files matching a pattern.
		Tries to exclude internal files automatically.
		Does not register temp-files.
		The default pattern registers everything (recursively, unless python 2).
		Returns which files were registered.
		"""
		from accelerator.extras import saved_files
		from glob import iglob
		pattern = os.path.normpath(pattern)
		assert not pattern.startswith('/')
		assert not pattern.startswith('../')
		forbidden = ('setup.json', 'post.json', 'method.tar.gz',)
		res = set()
		if PY3:
			files = iglob(pattern, recursive=True)
		else:
			# No recursive support on python 2.
			files = iglob(pattern)
		for fn in files:
			if (
				fn in forbidden or
				fn.startswith('DS/') or
				fn.startswith('OUTPUT/') or
				fn.startswith('Analysis.') or
				not os.path.isfile(fn)
			):
				continue
			key = self.filename(fn)
			# Don't override temp-ness of already registered files.
			if key not in saved_files:
				saved_files[key] = False
				res.add(fn)
		return res

	def input_filename(self, *parts):
		return os.path.join(self.input_directory, *parts)

	def open_input(self, filename, mode='r', encoding=None, errors=None):
		assert 'r' in mode, "Don't write to input files"
		if 'b' not in mode and encoding is None:
			encoding = 'utf-8'
		return open(self.input_filename(filename), mode, encoding=encoding, errors=errors)


class NoJob(Job):
	"""
	A empty string that is used for unset job arguments, with some properties
	that may still make sense on an unset job.
	Also provides .load() and .load_json() methods that return None as long
	as no filename or sliceno was specified, and .files() that always returns
	an empty set.
	"""

	__slots__ = ()

	# functions you shouldn't call on this
	filename = link_result = open = params = path = post = withfile = None

	workdir = None
	method = output = ''
	number = version = -1

	def __new__(cls):
		return unicode.__new__(cls, '')

	def dataset(self, name='default'):
		raise NoSuchDatasetError('NoJob has no datasets')

	@property
	def datasets(self):
		from accelerator.dataset import DatasetList
		return DatasetList()

	def files(self, pattern='*'):
		return set()

	def load(self, filename=None, sliceno=None, encoding='bytes', default=_nodefault):
		if default is not _nodefault:
			return default
		if filename is not None or sliceno is not None:
			raise NoSuchJobError('Can not load named / sliced file on <NoJob>')
		return None

	def json_load(self, filename=None, sliceno=None, unicode_as_utf8bytes=PY2, default=_nodefault):
		return self.load(filename, sliceno, default=default)

	@property # so it can return the same instance as all other NoJob things
	def parent(self):
		return NoJob

NoJob = NoJob()


class JobWithFile(namedtuple('JobWithFile', 'job name sliced extra')):
	__slots__ = ()

	def __new__(cls, job, name, sliced=False, extra=None):
		if isinstance(name, Path):
			name = str(name)
		assert not name.startswith('/'), "Specify relative filenames to JobWithFile"
		return tuple.__new__(cls, (Job(job), name, bool(sliced), extra,))

	def filename(self, sliceno=None):
		if sliceno is None:
			assert not self.sliced, "A sliced file requires a sliceno"
		else:
			assert self.sliced, "An unsliced file can not have a sliceno"
		return self.job.filename(self.name, sliceno)

	def load(self, sliceno=None, encoding='bytes', default=_nodefault):
		"""blob.load this file"""
		from accelerator.extras import pickle_load
		try:
			return pickle_load(self.filename(sliceno), encoding=encoding)
		except FileNotFoundError:
			if default is _nodefault:
				raise
			return default

	def json_load(self, sliceno=None, unicode_as_utf8bytes=PY2, default=_nodefault):
		from accelerator.extras import json_load
		try:
			return json_load(self.filename(sliceno), unicode_as_utf8bytes=unicode_as_utf8bytes)
		except FileNotFoundError:
			if default is _nodefault:
				raise
			return default

	def open(self, mode='r', sliceno=None, encoding=None, errors=None):
		return self.job.open(self.name, mode, sliceno, encoding, errors)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2020-2024 Carl Drougge                                     #
# Modifications copyright (c) 2024 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function

import bottle
import json
import sys
import os
import tarfile
import itertools
import collections
import functools
import mimetypes
import time
from stat import S_ISDIR, S_ISLNK

from accelerator.job import Job, JobWithFile
from accelerator.dataset import Dataset
from accelerator.unixhttp import call, WaitressServer
from accelerator.build import fmttime
from accelerator.configfile import resolve_listen
from accelerator.error import NoSuchWhateverError
from accelerator.metadata import insert_metadata
from accelerator.shell.parser import ArgumentParser, name2job, name2ds
from accelerator.shell.workdir import job_data, workdir_jids
from accelerator.compat import setproctitle, url_quote, urlencode
from accelerator.compat import FileNotFoundError, PermissionError
from accelerator import __version__ as ax_version
from accelerator import graph

# why wasn't Accept specified in a sane manner (like sending it in preference order)?
def get_best_accept(*want):
	d = {want[0]: -1} # fallback to first specified
	# {'a/*': 'a/exact'}, reversed() so earlier win
	want_short = {w.split('/', 1)[0] + '/*': w for w in reversed(want)}
	want = set(want)
	want.update(want_short)
	for accept in bottle.request.headers.get('Accept', '').split(','):
		accept = accept.split(';')
		mimetype = accept[0].strip()
		if mimetype in want:
			d[mimetype] = 1.0
			for p in accept[1:]:
				p = p.strip()
				if p.startswith('q='):
					try:
						d[mimetype] = float(p[2:])
					except ValueError:
						pass
	# include k in sort key as well so /* gets lower priority
	_, best = sorted(((v, k) for k, v in d.items()), reverse=True)[0]
	return want_short.get(best, best)

class JSONEncoderWithSet(json.JSONEncoder):
	__slots__ = ()
	def default(self, o):
		if isinstance(o, set):
			return list(o)
		return json.JSONEncoder.default(o)

json_enc = JSONEncoderWithSet(indent=4, ensure_ascii=False).encode


def ax_repr(o):
	res = []
	if isinstance(o, JobWithFile):
		link = '/job/' + bottle.html_escape(o.job)
		res.append('JobWithFile(<a href="%s">job=' % (link,))
		res.append(ax_repr(o.job))
		name = bottle.html_escape(o.name)
		if o.sliced:
			name += '.0'
		res.append('</a>, <a href="%s/%s">name=' % (link, name,))
		res.append(ax_repr(o.name))
		res.append('</a>, sliced=%s, extra=%s' % (ax_repr(o.sliced), ax_repr(o.extra),))
		res.append(')')
	elif isinstance(o, (list, tuple)):
		bra, ket = ('[', ']',) if isinstance(o, list) else ('(', ')',)
		res.append(bra)
		comma = ''
		for v in o:
			res.append(comma)
			res.append(ax_repr(v))
			comma = ', '
		res.append(ket)
	elif isinstance(o, dict):
		res.append('{')
		comma = ''
		for k, v in o.items():
			res.append(comma)
			res.append(ax_repr(k))
			res.append(': ')
			res.append(ax_repr(v))
			comma = ', '
		res.append('}')
	else:
		res.append(bottle.html_escape(repr(o)))
	return ''.join(res)

def ax_link(v):
	if isinstance(v, tuple):
		return '(%s)' % (', '.join(ax_link(vv) for vv in v),)
	elif isinstance(v, list):
		return '[%s]' % (', '.join(ax_link(vv) for vv in v),)
	elif v:
		ev = bottle.html_escape(v)
		if isinstance(v, Dataset):
			job = bottle.html_escape(v.job)
			name = bottle.html_escape(v.name)
			return '<a href="/job/%s">%s</a>/<a href="/dataset/%s">%s</a>' % (url_quote(v.job), job, url_quote(v), name,)
		elif isinstance(v, Job):
			return '<a href="/job/%s">%s</a>' % (url_quote(v), ev,)
		else:
			return ev
	else:
		return ''

_development = False

name2hashed = {}
hashed = {}

# Make contents-based names so that the files can be cached forever
def populate_hashed():
	from hashlib import sha1
	from base64 import b64encode
	dirname = os.path.join(os.path.dirname(__file__), 'board')
	for filename, ctype in [
		('style.css', 'text/css; charset=UTF-8'),
		('script.js', 'text/javascript; charset=UTF-8'),
		('graph.js', 'text/javascript; charset=UTF-8'),
	]:
		try:
			with open(os.path.join(dirname, filename), 'rb') as fh:
				data = fh.read()
			h = b64encode(sha1(data).digest(), b'_-').rstrip(b'=').decode('ascii')
			h_name = h + '/' + filename
			name2hashed[filename] = '/h/' + h_name
			hashed[h_name] = (data, ctype,)
		except OSError as e:
			name2hashed[filename] = '/h/ERROR'
			print(e, file=sys.stderr)


def js_quote(obj):
	# If the string contains '</script>' that will end the script, even if
	# it's in the middle of a string. Escape < to avoid problems like that.
	return json.dumps(obj).replace('<', '\\074')


def template(tpl_name, **kw):
	return bottle.template(
		tpl_name,
		ax_repr=ax_repr,
		ax_link=ax_link,
		ax_version=ax_version,
		js_quote=js_quote,
		name2hashed=name2hashed,
		template=template,
		url_quote=url_quote,
		**kw
	)


def view(name, subkey=None, prefer_ctype='application/json'):
	def view_decorator(func):
		@functools.wraps(func)
		def view_wrapper(**kw):
			if _development:
				populate_hashed()
			res = func(**kw)
			if isinstance(res, dict):
				accept = get_best_accept(prefer_ctype, 'application/json', 'text/json', 'text/html')
				if accept == 'text/html':
					return template(name, **res)
				elif prefer_ctype == accept == 'image/svg+xml':
					bottle.response.content_type = 'image/svg+xml; charset=UTF-8'
					return template(name, **res)
				else:
					bottle.response.content_type = accept + '; charset=UTF-8'
					if callable(subkey):
						res = subkey(res)
					elif subkey:
						res = res[subkey]
					return [json_enc(res), '\n']
			return res
		return view_wrapper
	return view_decorator

def fix_stacks(stacks, report_t):
	pid2pid = {}
	pid2jid = {}
	pid2part = {}
	job_pid = None
	for pid, indent, msg, t in stacks:
		if pid not in pid2pid and pid not in pid2jid:
			if msg.startswith('analysis('):
				pid2part[pid] = ''.join(c for c in msg if c.isdigit())
				pid2pid[pid] = job_pid
			else:
				pid2jid[pid] = msg.split(' ', 1)[0]
				job_pid = pid
		elif pid not in pid2part:
			pid2part[pid] = msg if msg in ('prepare', 'synthesis') else 'analysis'
		jobpid = pid
		while jobpid in pid2pid:
			jobpid = pid2pid[jobpid]
		jid = pid2jid[jobpid]
		if indent < 0:
			msg = msg.split('\n')
			start = len(msg) - 1
			while start and sum(map(bool, msg[start:])) < 5:
				start -= 1
			msg = [line.rstrip('\r') for line in msg[start:]]
			t = fmttime(report_t - t)
		else:
			t = fmttime(report_t - t, short=True)
		yield (jid, pid, indent, pid2part.get(pid), msg, t)

# datasets aren't dicts, so can't be usefully json encoded
def ds_json(d):
	ds = d['ds']
	keys = ('job', 'name', 'parent', 'filename', 'previous', 'hashlabel', 'lines')
	res = {k: getattr(ds, k) for k in keys}
	res['method'] = ds.job.method
	res['columns'] = {k: c.type for k, c in ds.columns.items()}
	return res

def main(argv, cfg):
	parser = ArgumentParser(prog=argv.pop(0), description='''runs a web server on listen_on (default localhost:8520, can be socket path) for displaying results (result_directory)''')
	parser.add_argument('--development', action='store_true', negation='not', help='enable tracebacks, disable caching, ...')
	parser.add_argument('listen_on', default='localhost:8520', nargs='?', help='host:port or path/to/socket')
	args = parser.parse_intermixed_args(argv)
	cfg.board_listen = resolve_listen(args.listen_on)[0]
	if isinstance(cfg.board_listen, str):
		# The listen path may be relative to the directory the user started us
		# from, but the reloader will exec us from the project directory, so we
		# have to be a little gross.
		cfg.board_listen = os.path.join(cfg.user_cwd, cfg.board_listen)
		argv[1:] = [cfg.board_listen]
	run(cfg, from_shell=True, development=args.development)

def run(cfg, from_shell=False, development=False):
	global _development
	_development = development

	project = os.path.split(cfg.project_directory)[1]
	setproctitle('ax board-server for %s on %s' % (project, cfg.board_listen,))

	# The default path filter (i.e. <something:path>) does not match newlines,
	# but we want it to do so (e.g. in case someone names a dataset with one).
	def pathfilter(config):
		return r'(?:.|\n)+?', None, None
	bottle.default_app[0].router.add_filter('path', pathfilter)

	populate_hashed()

	def call_s(*path, **kw):
		if kw:
			data = urlencode(kw).encode('utf-8')
		else:
			data = None
		return call(os.path.join(cfg.url, *map(url_quote, path)), data=data)

	def call_u(*path, **kw):
		url = os.path.join(cfg.urd, *map(url_quote, path))
		if kw:
			url = url + '?' + urlencode(kw)
		return call(url, server_name='urd')

	# yield chunks of a file, optionally inserting some bonus content somewhere
	def file_parts_iter(parts, clen, fh, start=0, end=None, chunksize=128 * 1024):
		try:
			if end is None:
				end = clen
			pos = 0
			for part, part_len in parts:
				if end <= pos:
					break
				if pos + part_len > start:
					if isinstance(part, bytes):
						yield part[max(0, start - pos):max(0, end - pos)]
					else:
						skip = max(0, start - pos)
						fh.seek(part + skip)
						z = part_len - skip - max(0, pos + part_len - end)
						while z > 0:
							want = min(z, chunksize)
							data = fh.read(want)
							yield data
							if len(data) != want:
								return # File shrunk? Nothing we can do.
							z -= chunksize
				pos += part_len
		finally:
			fh.close()

	# Based on the one in bottle but modified for our needs.
	def static_file(filename, root, job=None):
		root = os.path.abspath(root) + os.sep
		filename = os.path.abspath(os.path.join(root, filename))
		if not filename.startswith(root):
			return bottle.HTTPError(403, "Access denied.")
		try:
			fh = open(filename, 'rb')
		except PermissionError:
			return bottle.HTTPError(403, "Access denied.")
		except FileNotFoundError:
			return bottle.HTTPError(404, "File does not exist.")
		headers = {}
		mimetype, encoding = mimetypes.guess_type(filename)
		if mimetype:
			if mimetype.startswith('text/'):
				mimetype += '; charset=UTF-8'
			headers['Content-Type'] = mimetype
		if encoding:
			headers['Content-Encoding'] = encoding
		stats = os.fstat(fh.fileno())
		lm = time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime(stats.st_mtime))
		headers['Last-Modified'] = lm

		ims = bottle.request.environ.get('HTTP_IF_MODIFIED_SINCE')
		if ims:
			ims = bottle.parse_date(ims.split(';')[0].strip())
		if ims is not None and ims >= int(stats.st_mtime):
			headers['Date'] = time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime())
			fh.close()
			return bottle.HTTPResponse(status=304, **headers)

		file_parts = insert_metadata(filename, fh, job, stats.st_size)
		clen = sum(z for _, z in file_parts)
		headers['Content-Length'] = clen

		if bottle.request.method == 'HEAD':
			body = ''
			fh.close()
		else:
			body = fh

		headers['Accept-Ranges'] = 'bytes'
		if 'HTTP_RANGE' in bottle.request.environ:
			ranges = list(bottle.parse_range_header(bottle.request.environ['HTTP_RANGE'], clen))
			if not ranges:
				return bottle.HTTPError(416, "Requested Range Not Satisfiable")
			offset, end = ranges[0]
			headers['Content-Range'] = 'bytes %d-%d/%d' % (offset, end-1, clen)
			headers['Content-Length'] = str(end-offset)
			if body:
				body = file_parts_iter(file_parts, clen, body, offset, end)
			return bottle.HTTPResponse(body, status=206, **headers)
		if body:
			body = file_parts_iter(file_parts, clen, body)
		return bottle.HTTPResponse(body, **headers)

	@bottle.get('/')
	@view('main')
	def main_page(path='/results'):
		return dict(
			project=project,
			workdirs=cfg.workdirs,
			path=path,
			url_path=url_quote(path),
		)

	# Look for actual workdirs, so things like /workdirs/foo/foo-37/foo-1/bar
	# resolves to ('foo-37', 'foo-1/bar') and not ('foo-1', 'bar').
	path2wd = {v: k for k, v in cfg.workdirs.items()}
	def job_and_file(path, default_name):
		wd = ''
		path = iter(path.split('/'))
		for name in path:
			if not name:
				continue
			wd = wd + '/' + name
			if wd in path2wd:
				break
		else:
			return None, default_name
		try:
			jobid = Job(next(path))
		except (StopIteration, NoSuchWhateverError):
			return None, default_name
		return jobid, '/'.join(path) or default_name

	def results_contents(path):
		files = {}
		dirs = {}
		res = {'files': files, 'dirs': dirs}
		default_jobid = None
		default_prefix = ''
		prefix = cfg.result_directory
		for part in path.strip('/').split('/'):
			prefix = os.path.join(prefix, part)
			if not default_jobid:
				try:
					default_jobid, default_prefix = job_and_file(os.readlink(prefix), '')
					if default_jobid and default_prefix:
						default_prefix += '/'
				except OSError:
					pass
			elif default_prefix:
				default_prefix += part + '/'
		filenames = os.listdir(prefix)
		for fn in filenames:
			if fn.startswith('.') or fn.endswith('_'):
				continue
			ffn = os.path.join(prefix, fn)
			try:
				lstat = os.lstat(ffn)
				if S_ISLNK(lstat.st_mode):
					link_dest = os.readlink(ffn)
					stat = os.stat(link_dest)
					jobid, name = job_and_file(link_dest, fn)
				else:
					stat = lstat
					jobid = default_jobid
					name = default_prefix + fn
			except OSError:
				continue
			if S_ISDIR(stat.st_mode):
				dirs[fn] = os.path.join('/results', path, fn, '')
			else:
				files[fn] = dict(
					jobid=jobid,
					name=name,
					ts=lstat.st_mtime,
					size=stat.st_size,
				)
		if path:
			a, b = os.path.split(path)
			dirs['..'] = os.path.join('/results', a, '') if a else '/'
		return res

	@bottle.get('/results')
	@bottle.get('/results/')
	@bottle.get('/results/<path:path>')
	def results(path=''):
		path = path.strip('/')
		if os.path.isdir(os.path.join(cfg.result_directory, path)):
			accept = get_best_accept('text/html', 'application/json', 'text/json')
			if accept == 'text/html':
				return main_page(path=os.path.join('/results', path).rstrip('/'))
			else:
				bottle.response.content_type = accept + '; charset=UTF-8'
				bottle.response.set_header('Cache-Control', 'no-cache')
				return json.dumps(results_contents(path))
		elif path:
			try:
				link_dest = os.readlink(os.path.join(cfg.result_directory, path))
				job, _ = job_and_file(link_dest, None)
			except OSError:
				job = None
			return static_file(path, root=cfg.result_directory, job=job)
		else:
			return {'missing': 'result directory %r missing' % (cfg.result_directory,)}

	@bottle.get('/status')
	@view('status')
	def status():
		status = call_s('status/full')
		if 'short' in bottle.request.query:
			if status.idle:
				return 'idle'
			else:
				t, msg, _ = status.current
				return '%s (%s)' % (msg, fmttime(status.report_t - t, short=True),)
		else:
			status.tree = list(fix_stacks(status.pop('status_stacks', ()), status.report_t))
			return status

	@bottle.get('/last_error')
	@view('last_error')
	def last_error():
		return call_s('last_error')

	@bottle.get('/job/<jobid>/method.tar.gz/')
	@bottle.get('/job/<jobid>/method.tar.gz/<name:path>')
	def job_method(jobid, name=None):
		job = name2job(cfg, jobid)
		with tarfile.open(job.filename('method.tar.gz'), 'r:gz') as tar:
			if name:
				info = tar.getmember(name)
			else:
				members = [info for info in tar.getmembers() if info.isfile()]
				if len(members) == 1 and not name:
					info = members[0]
				else:
					return template('job_method_list', members=members, job=job)
			bottle.response.content_type = 'text/plain; charset=UTF-8'
			return tar.extractfile(info).read()

	@bottle.get('/job/<jobid>/<name:path>')
	def job_file(jobid, name):
		job = name2job(cfg, jobid)
		res = static_file(name, root=job.path, job=job)
		if not res.content_type and res.status_code < 400:
			# bottle default is text/html, which is probably wrong.
			res.content_type = 'text/plain'
		return res

	@bottle.get('/job/<jobid>')
	@bottle.get('/job/<jobid>/')
	@view('job')
	def job(jobid):
		job = name2job(cfg, jobid)
		try:
			post = job.post
		except IOError:
			post = None
		if post:
			aborted = False
			files = [fn for fn in job.files() if fn[0] != '/']
			jobs = list(post.subjobs)
			jobs.append(job)
			jobs = call_s('jobs_are_current', jobs='\0'.join(jobs))
			subjobs = [(Job(jobid), jobs[jobid]) for jobid in post.subjobs]
			current = jobs[job]
		else:
			aborted = True
			current = False
			files = None
			subjobs = None
		return dict(
			job=job,
			aborted=aborted,
			current=current,
			output=os.path.exists(job.filename('OUTPUT')),
			datasets=job.datasets,
			params=job.params,
			subjobs=subjobs,
			files=files,
		)

	@bottle.get('/dataset/<dsid:path>')
	@view('dataset', ds_json)
	def dataset(dsid):
		ds = name2ds(cfg, dsid.rstrip('/'))
		q = bottle.request.query
		if q.column:
			lines = int(q.lines or 10)
			it = ds.iterate(None, q.column)
			it = itertools.islice(it, lines)
			t = ds.columns[q.column].type
			if t in ('datetime', 'date', 'time',):
				it = map(str, it)
			elif t in ('bytes', 'pickle',):
				it = map(repr, it)
			res = list(it)
			bottle.response.content_type = 'application/json; charset=UTF-8'
			return json.dumps(res)
		else:
			return dict(ds=ds)

	@bottle.get('/graph/job/<jobid>')
	@view('rendergraph', prefer_ctype='image/svg+xml')
	def job_graph(jobid):
		job = name2job(cfg, jobid)
		ret = graph.graph(job, 'job')
		return ret

	@bottle.get('/graph/dataset/<dsid:path>')
	@view('rendergraph', prefer_ctype='image/svg+xml')
	def dataset_graph(dsid):
		ds = name2ds(cfg, dsid.rstrip('/'))
		ret = graph.graph(ds, 'dataset')
		return ret

	@bottle.get('/graph/urd/<user>/<build>/<ts>')
	@view('rendergraph', prefer_ctype='image/svg+xml')
	def urd_graph(user, build, ts):
		key = user + '/' + build + '/' + ts
		d = call_u(key)
		ret = graph.graph(d, 'urd')
		return ret

	def load_workdir(jobs, name):
		known = call_s('workdir', name)
		for jid in workdir_jids(cfg, name):
			jobs[jid] = job_data(known, jid)
		return jobs

	@bottle.get('/workdir/<name>')
	@view('workdir', 'jobs')
	def workdir(name):
		return dict(name=name, jobs=load_workdir(collections.OrderedDict(), name))

	@bottle.get('/workdir')
	@bottle.get('/workdir/')
	@view('workdir', 'jobs')
	def all_workdirs():
		jobs = collections.OrderedDict()
		for name in sorted(cfg.workdirs):
			load_workdir(jobs, name)
		return dict(name='ALL', jobs=jobs)

	@bottle.get('/methods')
	@view('methods')
	def methods():
		methods = call_s('methods')
		by_package = collections.defaultdict(list)
		for name, data in sorted(methods.items()):
			by_package[data.package].append(name)
		by_package.pop('accelerator.test_methods', None)
		return dict(methods=methods, by_package=by_package)

	@bottle.get('/method/<name>')
	@view('method', 'data')
	def method(name):
		methods = call_s('methods')
		if name not in methods:
			return bottle.HTTPError(404, 'Method %s not found' % (name,))
		return dict(name=name, data=methods[name], cfg=cfg)

	@bottle.get('/urd')
	@bottle.get('/urd/')
	@view('urd', 'lists')
	def urd():
		return dict(
			lists=call_u('list'),
			project=project,
		)

	@bottle.get('/urd/<user>/<build>')
	@bottle.get('/urd/<user>/<build>/')
	@view('urdlist', 'timestamps')
	def urdlist(user, build):
		key = user + '/' + build
		return dict(
			key=key,
			timestamps=call_u(key, 'since/0', captions=1),
		)

	@bottle.get('/urd/<user>/<build>/<ts>')
	@view('urditem', 'entry')
	def urditem(user, build, ts):
		key = user + '/' + build + '/' + ts
		d = call_u(key)
		return dict(key=key, entry=d)

	@bottle.get('/h/<name:path>')
	def hashed_file(name):
		if name not in hashed:
			return bottle.HTTPError(404, 'Not Found')
		data, ctype = hashed[name]
		bottle.response.content_type = ctype
		bottle.response.set_header('Cache-Control', 'max-age=604800, immutable')
		return data

	@bottle.get('/robots.txt')
	def robots_txt():
		bottle.response.content_type = 'text/plain'
		return b'User-agent: *\nDisallow: /\n'

	@bottle.error(500)
	def error(e):
		tpl = bottle.ERROR_PAGE_TEMPLATE
		if isinstance(e.exception, NoSuchWhateverError):
			e.body = str(e.exception)
		return bottle.template(tpl, e=e)

	bottle.TEMPLATE_PATH = [os.path.join(os.path.dirname(__file__), 'board')]
	kw = {}
	if development:
		kw['reloader'] = True
		kw['debug'] = True
	if not from_shell:
		kw['quiet'] = True
	kw['server'] = WaitressServer
	listen = cfg.board_listen
	if isinstance(listen, tuple):
		kw['host'], kw['port'] = listen
	else:
		from accelerator.server import check_socket
		check_socket(listen)
		kw['host'] = listen
		kw['port'] = 0
	bottle.run(**kw)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2023-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

description = """
Test using the returned object from job.save() and job.json_save()
"""

from pathlib import Path


def save(job, name, sliceno):
	p = job.save('contents of %s %s' % (name, sliceno,), name + '.pickle', sliceno=sliceno)
	j = job.json_save({name: sliceno}, name + '.json', sliceno=sliceno)
	name += '_path'
	p_path = job.save('contents of %s %s' % (name, sliceno,), Path(name + '.pickle'), sliceno=sliceno)
	j_path = job.json_save({name: sliceno}, Path(name + '.json'), sliceno=sliceno)
	return p, j, p_path, j_path

def check(job, name, sliceno, p, j, p_path, j_path):
	assert p.load() == job.load(Path(p.filename)) == 'contents of %s %s' % (name, sliceno,)
	assert j.load() == job.json_load(Path(j.filename)) == {name: sliceno}
	assert p_path.load() == job.load(Path(p_path.filename)) == 'contents of %s_path %s' % (name, sliceno,)
	assert j_path.load() == job.json_load(Path(j_path.filename)) == {name + '_path': sliceno}
	for obj, filename in [
		# Use Path() for files saved with strings and strings for files saved with Path.
		(p, Path(name + '.pickle')),
		(j, Path(name + '.json')),
		(p_path, name + '_path.pickle'),
		(j_path, name + '_path.json'),
	]:
		path = job.filename(filename, sliceno=sliceno)
		assert obj.path == path
		assert obj.filename == path.split('/')[-1]
		assert obj.jobwithfile().filename(sliceno) == path

def prepare(job):
	res = save(job, 'prepare', None)
	check(job, 'prepare', None, *res)
	return res

def analysis(sliceno, job, prepare_res):
	check(job, 'prepare', None, *prepare_res)
	res = save(job, 'analysis', sliceno)
	check(job, 'analysis', sliceno, *res)
	return res

def synthesis(job, prepare_res, analysis_res):
	check(job, 'prepare', None, *prepare_res)
	for sliceno, res in enumerate(analysis_res):
		check(job, 'analysis', sliceno, *res)
	res = save(job, 'synthesis', None)
	check(job, 'synthesis', None, *res)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2024 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import gzip
import math
import os
import struct
import sys

from accelerator._dsutil import siphash24
from accelerator.dsutil import typed_writer
from accelerator.subjobs import build

description = r'''
Test NaN values in the various float supporting types, making sure they
hash as expected, including for unusual NaN representations.
'''

def synthesis(job, slices):
	normal_NaN = float('NaN')
	unusual_NaN_0 = struct.unpack('=f', b'\x7f\xc0\xc0\x7f')[0]
	unusual_NaN_1 = struct.unpack('=f', b'\x7f\xc1\xc1\x7f')[0]
	assert math.isnan(normal_NaN)
	assert math.isnan(unusual_NaN_0)
	assert math.isnan(unusual_NaN_1)
	assert struct.pack('=f', unusual_NaN_0) == b'\x7f\xc0\xc0\x7f'
	assert struct.pack('=f', unusual_NaN_1) == b'\x7f\xc1\xc1\x7f'

	f32_normal_NaN = struct.pack('=f', normal_NaN)
	f32_unusual_NaN_0 = struct.pack('=f', unusual_NaN_0)
	f32_unusual_NaN_1 = struct.pack('=f', unusual_NaN_1)
	f64_normal_NaN = struct.pack('=d', normal_NaN)
	f64_unusual_NaN_0 = struct.pack('=d', unusual_NaN_0)
	f64_unusual_NaN_1 = struct.pack('=d', unusual_NaN_1)

	if sys.byteorder == 'big':
		f32_std_nan = b'\x7f\xc0\x00\x00'
		f64_std_nan = b'\x7f\xf8\x00\x00\x00\x00\x00\x00'
	else:
		f32_std_nan = b'\x00\x00\xc0\x7f'
		f64_std_nan = b'\x00\x00\x00\x00\x00\x00\xf8\x7f'

	# This is always what should be hashed, regardless of platform NaN representation.
	want_hash_float = siphash24(f64_std_nan)
	want_hash_complex = siphash24(f64_std_nan + f64_std_nan)

	def test_encoding_and_hash(typ, nan, *want):
		writer = typed_writer(typ)
		assert writer.hash(nan) == want_hash_float
		with writer('tmp', compression='gzip', none_support=True) as w:
			w.write(nan)
		with gzip.open('tmp', 'rb') as fh:
			got_bytes = fh.read()
		os.remove('tmp')
		assert got_bytes in want, 'bad NaN representation in %s: got %r, wanted something in %r' % (typ, got_bytes, want,)

	# Test that they either encode as themselves or as one of the normal NaNs,
	# and all hash the same as the standard float64 NaN.
	test_encoding_and_hash('float32', normal_NaN, f32_std_nan, f32_normal_NaN)
	test_encoding_and_hash('float32', unusual_NaN_0, f32_std_nan, f32_normal_NaN, f32_unusual_NaN_0)
	test_encoding_and_hash('float32', unusual_NaN_1, f32_std_nan, f32_normal_NaN, f32_unusual_NaN_1)
	test_encoding_and_hash('float64', normal_NaN, f64_std_nan, f64_normal_NaN)
	test_encoding_and_hash('float64', unusual_NaN_0, f64_std_nan, f64_normal_NaN, f64_unusual_NaN_0)
	test_encoding_and_hash('float64', unusual_NaN_1, f64_std_nan, f64_normal_NaN, f64_unusual_NaN_1)
	test_encoding_and_hash('number', normal_NaN, b'\x01' + f64_std_nan, b'\x01' + f64_normal_NaN)
	test_encoding_and_hash('number', unusual_NaN_0, b'\x01' + f64_std_nan, b'\x01' + f64_normal_NaN, b'\x01' + f64_unusual_NaN_0)
	test_encoding_and_hash('number', unusual_NaN_1, b'\x01' + f64_std_nan, b'\x01' + f64_normal_NaN, b'\x01' + f64_unusual_NaN_1)

	def mk_dws(typ):
		kw = {
			'name': typ + ' NaNs',
			'columns': {'ix': 'ascii', 'nan': typ},
		}
		dw_unhashed = job.datasetwriter(**kw)
		kw['name'] += ' hashed'
		kw['hashlabel'] = 'nan'
		dw_hashed = job.datasetwriter(**kw)
		return dw_unhashed, dw_unhashed.get_split_write(), dw_hashed, dw_hashed.get_split_write()

	for typ, values in (
		('complex32', [complex(normal_NaN, normal_NaN), complex(unusual_NaN_0, unusual_NaN_1)]),
		('complex64', [complex(normal_NaN, normal_NaN), complex(unusual_NaN_1, unusual_NaN_0)]),
		('float32', [normal_NaN, unusual_NaN_0, unusual_NaN_1]),
		('float64', [normal_NaN, unusual_NaN_0, unusual_NaN_1]),
		('number', [normal_NaN, unusual_NaN_0, unusual_NaN_1]),
	):
		dw_u, w_u, dw_h, w_h = mk_dws(typ)
		h = typed_writer(typ).hash
		want_h = (want_hash_complex if typ.startswith('complex') else want_hash_float)
		for ix, v in enumerate(values):
			assert want_h == h(v), 'value index %d did not hash correctly for type %s' % (ix, typ,)
			w_u(str(ix), v)
			w_h(str(ix), v)
		ds_h = dw_h.finish()
		assert set(ds_h.lines) == {0, len(values)}, 'Not all NaNs ended up in the same slice in %s' % (ds_h.quoted,)
		expect_lines = ds_h.lines
		ds_u = dw_u.finish()
		ds = build('dataset_hashpart', source=ds_u, hashlabel='nan').dataset()
		assert set(ds.lines) == {0, len(values)}, 'Not all NaNs ended up in the same slice in %s (dataset_hashpart from %s)' % (ds.quoted, ds_u.quoted,)
		assert expect_lines == ds.lines, 'dataset_hashpart (%s) disagrees with datasetwriter (%s) about NaN slicing' % (ds.quoted, ds_h.quoted,)
		ds = build('dataset_type', source=ds_u, hashlabel='nan', column2type={'ix': 'number'}).dataset()
		assert set(ds.lines) == {0, len(values)}, 'Not all NaNs ended up in the same slice in %s (dataset_type from %s)' % (ds.quoted, ds_u.quoted,)
		assert expect_lines == ds.lines, 'dataset_type (%s) disagrees with datasetwriter (%s) about NaN slicing' % (ds.quoted, ds_h.quoted,)
		rehash_lines = [len(list(ds_u.iterate(sliceno, rehash=True, hashlabel='nan'))) for sliceno in range(slices)]
		assert expect_lines == rehash_lines, 'ds.iterate(rehash=True) of %s disagrees with datasetwriter hashing (%s) about NaN slicing' % (ds_u.quoted, ds_h.quoted,)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2019-2024 Carl Drougge                                     #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify the dataset_hashpart method with various options.
'''

from datetime import date

from accelerator import subjobs
from accelerator.dataset import DatasetWriter, Dataset
from accelerator.dsutil import typed_writer

data = [
	{"a column": "text", "also a column": b"octets", "number": 10, "date": date(1978, 1, 1)},
	{"a column": "a", "also a column": b"0", "number": -900, "date": date(2009, 1, 1)},
	{"a column": "b", "also a column": b"1", "number": -901, "date": date(2009, 1, 2)},
	{"a column": "c", "also a column": b"2", "number": -902, "date": date(2009, 1, 3)},
	{"a column": "d", "also a column": b"3", "number": -903, "date": date(2009, 1, 4)},
	{"a column": "e", "also a column": b"4", "number": -904, "date": date(2009, 1, 5)},
	{"a column": "f", "also a column": b"5", "number": -905, "date": date(2009, 1, 6)},
	{"a column": "g", "also a column": b"6", "number": -906, "date": date(2009, 1, 7)},
	{"a column": "h", "also a column": b"7", "number": -907, "date": date(2009, 1, 8)},
	{"a column": "i", "also a column": b"8", "number": -908, "date": date(2009, 1, 9)},
	{"a column": "j", "also a column": b"9", "number": -909, "date": date(2009, 2, 1)},
	{"a column": "k", "also a column": b"Z", "number": -999, "date": date(2009, 3, 1)},
	{"a column": "l", "also a column": b"Y", "number": -989, "date": date(2009, 4, 1)},
	{"a column": "m", "also a column": b"X", "number": -979, "date": date(2009, 5, 1)},
	{"a column": "n", "also a column": b"W", "number": -969, "date": date(2009, 6, 1)},
	{"a column": "o", "also a column": b"V", "number": -959, "date": date(2009, 7, 1)},
	{"a column": "p", "also a column": b"A", "number": -949, "date": date(2009, 8, 1)},
	{"a column": "q", "also a column": b"B", "number": -939, "date": date(2009, 9, 1)},
	{"a column": "r", "also a column": b"C", "number": -929, "date": None},
	{"a column": "s", "also a column": b"D", "number": None, "date": date(1970, 1, 1)},
	{"a column": "B", "also a column": None, "number": -242, "date": date(1970, 2, 3)},
	{"a column": None, "also a column": b"F", "number": -123, "date": date(1970, 4, 1)},
]
bonus_data = [
	{"a column": "foo", "also a column": b"bar", "number": 42, "date": date(2019, 4, 10)},
]
columns = {
	"a column": ("ascii", True),
	"also a column": ("bytes", True),
	"number": ("int32", True),
	"date": ("date", True),
}

def write(data, columns=columns, **kw):
	dw = DatasetWriter(columns=columns, **kw)
	w = dw.get_split_write_dict()
	for values in data:
		w(values)
	return dw.finish()

def verify(slices, data, source, previous=None, **options):
	jid = subjobs.build(
		"dataset_hashpart",
		datasets=dict(source=source, previous=previous),
		options=options,
	)
	hl = options["hashlabel"]
	h = typed_writer(columns[hl][0]).hash
	ds = Dataset(jid)
	good = {row[hl]: row for row in data}
	names = list(source.columns)
	for slice in range(slices):
		for row in ds.iterate_chain(slice, names):
			row = dict(zip(names, row))
			assert h(row[hl]) % slices == slice, "row %r is incorrectly in slice %d in %s" % (row, slice, ds)
			want = good[row[hl]]
			assert row == want, '%s (rehashed from %s) did not contain the right data for "%s".\nWanted\n%r\ngot\n%r' % (ds, source, hl, want, row)
	want_lines = len(data)
	got_lines = ds.chain().lines()
	assert got_lines == want_lines, '%s (rehashed from %s) had %d lines, should have had %d' % (ds, source, got_lines, want_lines,)
	return ds

def verify_empty(source, previous=None, **options):
	jid = subjobs.build(
		"dataset_hashpart",
		datasets=dict(source=source, previous=previous),
		options=options,
	)
	ds = Dataset(jid)
	chain = ds.chain_within_job()
	assert list(chain.iterate(None)) == [], "source=%s previous=%s did not produce empty dataset in %s" % (source, previous, ds,)
	assert chain[0].previous == previous, "Empty %s should have had previous=%s, but had %s" % (ds, previous, chain[0].previous,)

def synthesis(params):
	ds = write(data)
	for colname in data[0]:
		verify(params.slices, data, ds, hashlabel=colname)
	# ok, all the hashing stuff works out, let's test the chaining options.
	bonus_ds = write(bonus_data, name="bonus", previous=ds)
	# no chaining options - full chain
	verify(params.slices, data + bonus_data, bonus_ds, hashlabel="date")
	# just the bonus ds
	verify(params.slices, bonus_data, bonus_ds, hashlabel="date", length=1)
	# built as a chain
	verify(params.slices, data + bonus_data, bonus_ds, hashlabel="date", chain_slices=True)
	# normal chaining
	a = verify(params.slices, data, ds, hashlabel="date")
	b = verify(params.slices, data + bonus_data, bonus_ds, hashlabel="date", previous=a)
	assert b.chain() == [a, b], "chain of %s is not [%s, %s] as expected" % (b, a, b)
	# chain_slices sparseness
	empty = write([], name="empty")
	verify_empty(empty, hashlabel="date")
	verify_empty(empty, hashlabel="date", chain_slices=True)
	ds = verify(params.slices, [], empty, hashlabel="date", chain_slices=True)
	# two populated slices with the same data, should end up in two datasets.
	dw = DatasetWriter(columns=columns, name="0 and 2", allow_missing_slices=True)
	dw.set_slice(0)
	dw.write_dict(data[0])
	dw.set_slice(2)
	dw.write_dict(data[0])
	ds = verify(params.slices, [data[0], data[0]], dw.finish(), hashlabel="date", chain_slices=True)
	got_slices = len(ds.chain())
	assert got_slices == params.slices, "%s (built with chain_slices=True) has %d datasets in chain, expected %d." % (ds, got_slices, params.slices)

	# test varying types and available columns over the chain (including the hashlabel type)
	v1 = write([{'a': '101', 'b':  201 }], columns={'a': 'ascii',  'b': 'int32'}, name='varying1')
	v2 = write([{'a':  102 , 'c': '202'}], columns={'a': 'number', 'c': 'ascii'}, name='varying2', previous=v1)
	v3 = write([{'a':  103             }], columns={'a': 'int32'               }, name='varying3', previous=v2)
	hashed_varying = subjobs.build("dataset_hashpart", source=v3, hashlabel="a").dataset().chain()
	assert len(hashed_varying) == 3
	for unhashed, hashed in zip([v1, v2, v3], hashed_varying):
		assert {n: dc.type for n, dc in unhashed.columns.items()} == {n: dc.type for n, dc in hashed.columns.items()}
		assert list(unhashed.iterate(0)) == list(hashed.iterate(None))
		assert hashed.hashlabel == 'a'
		hash_t = unhashed.columns['a'].type
		hash_v = next(unhashed.iterate(0, 'a'))
		want_slice = typed_writer(hash_t).hash(hash_v) % params.slices
		assert hashed.lines[want_slice] == 1



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2019-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify the zip wrapper for csvimport.
'''

from zipfile import ZipFile
from io import BytesIO
from gzip import GzipFile
from collections import OrderedDict

from accelerator import subjobs
from accelerator.dataset import Dataset
from accelerator import g

file_a = b'0\nfoo\n'
list_a = [b'foo']
file_b = b'0\nbar\n'
list_b = [b'bar']

def compress(data):
	# In python3 I could just use gzip.compress..
	with BytesIO() as bio:
		with GzipFile(fileobj=bio, mode='wb') as fh:
			fh.write(data)
		return bio.getvalue()

def verify(zipname, inside_filenames, want_ds, **kw):
	opts=dict(
		filename=g.job.filename(zipname),
		inside_filenames=inside_filenames,
	)
	opts.update(kw)
	jid = subjobs.build('csvimport_zip', options=opts)
	for dsn, want_data in want_ds.items():
		got_data = list(Dataset(jid, dsn).iterate(None, '0'))
		assert got_data == want_data, "%s/%s from %s didn't contain %r, instead contained %r" % (jid, dsn, zipname, want_data, got_data)

def verify_order(want_order, namemap={}, **kw):
	opts=dict(
		filename=g.job.filename('many files.zip'),
		label_lines=0,
		labels=['0'],
	)
	opts.update(kw)
	jid = subjobs.build('csvimport_zip', options=opts)
	want_order = list(want_order)
	for dsn in want_order:
		b = b'contents of ' + namemap.get(dsn, dsn).encode('ascii')
		got_data = list(Dataset(jid, dsn).iterate(None, '0'))
		assert got_data == [b], "%s/%s from 'many files.zip' didn't contain [%r], instead contained %r" % (jid, dsn, b, got_data)
	got_order = [ds.name for ds in Dataset(jid, want_order[-1]).chain()]
	assert want_order == got_order, 'Wanted order %r, got %r in %s' % (want_order, got_order, jid,)
	if len(want_order) > 1: # chaining is actually on, so a default ds is produced
		got_order = [ds.name for ds in Dataset(jid).chain()]
		want_order[-1] = 'default'
		assert want_order == got_order, 'Wanted order %r, got %r in %s' % (want_order, got_order, jid,)

def synthesis():
	# Simple case, a single file in the zip.
	with ZipFile('a.zip', 'w') as z:
		z.writestr('a', file_a)
	with ZipFile('b.zip', 'w') as z:
		z.writestr('b', file_b)
	# Several files.
	with ZipFile('both.zip', 'w') as z:
		z.writestr('a', file_a)
		z.writestr('b', file_b)
	# With one of the files gziped.
	with ZipFile('both, b compressed.zip', 'w') as z:
		z.writestr('a', file_a)
		z.writestr('b', compress(file_b))
	# Repeated name inside.
	with ZipFile('both called a.zip', 'w') as z:
		z.writestr('a', file_a)
		z.writestr('a', file_b)
	# Same, with one compressed.
	with ZipFile('both called a, first compressed.zip', 'w') as z:
		z.writestr('a', compress(file_a))
		z.writestr('a', file_b)
	# A whole bunch of files.
	with ZipFile('many files.zip', 'w') as z:
		manyfiles = {}
		for a in 'BACDEFGHIJKLMOPQRSTUVWXYZ': # almost in order
			b = b'contents of ' + a.encode('ascii')
			z.writestr(a, b)
			manyfiles[a] = [b]
	# Make sure having a file named "default" doesn't cause issues.
	with ZipFile('named default.zip', 'w') as z:
		z.writestr('default', file_b)
	verify('a.zip', {'a': 'foo'}, {'foo': list_a, 'default': list_a})
	verify('a.zip', {}, {'a': list_a, 'default': list_a})
	verify('b.zip', {'b': 'bar'}, {'bar': list_b, 'default': list_b})
	verify('b.zip', {}, {'b': list_b, 'default': list_b})
	verify('both.zip', {'a': 'foo', 'b': 'bar'}, {'foo': list_a, 'bar': list_b})
	verify('both.zip', {}, {'a': list_a, 'b': list_b})
	verify('both.zip', {'a': 'foo'}, {'foo': list_a, 'default': list_a})
	verify('both, b compressed.zip', {'a': 'foo', 'b': 'bar'}, {'foo': list_a, 'bar': list_b})
	verify('both, b compressed.zip', {}, {'a': list_a, 'b': list_b})
	verify('both called a.zip', {}, {'a': list_a, 'a_': list_b})
	verify('both called a, first compressed.zip', {}, {'a': list_a, 'a_': list_b})
	verify('many files.zip', {}, manyfiles, label_lines=0, labels=['0'])
	verify('named default.zip', {}, {'default': list_b})
	# Use inside_filenames to test this again in a different way.
	verify('a.zip', {'a': 'default'}, {'default': list_a})

	# check all the different orders chaining can be done in
	verify_order('BACDEFGHIJKLMOPQRSTUVWXYZ') # implicitly 'on', i.e. 'by_ziporder'
	inside_filenames = OrderedDict([
		('A', 'A'),
		('Y', 'B'),
		('B', 'C'),
		('E', 'E'),
		('D', 'D'),
	])
	verify_order('ABCDE', chaining='by_dsname'  , inside_filenames=inside_filenames, namemap={'C': 'B', 'B': 'Y'})
	verify_order('ACDEB', chaining='by_filename', inside_filenames=inside_filenames, namemap={'C': 'B', 'B': 'Y'})
	verify_order('CADEB', chaining='by_ziporder', inside_filenames=inside_filenames, namemap={'C': 'B', 'B': 'Y'})
	verify_order('ABCED', chaining='by_dict'    , inside_filenames=inside_filenames, namemap={'C': 'B', 'B': 'Y'})
	# implicitly 'on', i.e. 'by_dict' since inside_filenames is set
	verify_order('ABCED',                         inside_filenames=inside_filenames, namemap={'C': 'B', 'B': 'Y'})
	# off doesn't chain, so each ds is a chain of length 1
	for dsn in inside_filenames.values():
		verify_order(dsn, chaining='off', inside_filenames=inside_filenames, namemap={'C': 'B', 'B': 'Y'})



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2024 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import os

description = r'''
Test that the expected files are registered (and not registered) when
using job.open(), job.register_file() and job.register_files().

This one registers nothing, but produces temp files, and a subdirectory.
'''

def analysis(sliceno, job):
	if sliceno == 0:
		with open('analysis slice 0.txt', 'w') as fh:
			fh.write('written in analysis slice 0')
	if sliceno == 1:
		with job.open('temporary analysis slice 1.txt', 'w', temp=True) as fh:
			fh.write('written as temporary in analysis slice 1')

def synthesis(job):
	with open('synthesis file.txt', 'w') as fh:
		fh.write('written in synthesis')
	with job.open('temporary synthesis file.txt', 'w', temp=True) as fh:
		fh.write('written as temporary in synthesis')
	os.mkdir('subdir')
	with open('subdir/file.txt', 'w') as fh:
		fh.write('written in a subdir, not automatically registered.')



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2024 Anders Berkeman                                       #
# Modifications copyright (c) 2024 Carl Drougge                            #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from math import sin, pi, tan, atan
from collections import defaultdict
from datetime import datetime
from accelerator import JobWithFile, Job
from accelerator.dataset import Dataset
from accelerator.compat import unicode, FileNotFoundError

MAXANGLE = 45 * pi / 180


def expand_to_set(what, fun=lambda x: x):
	if what:
		if not isinstance(what, (list, tuple)):
			what = [what, ]
		return set(fun(item) for item in what)
	return set()


def jobdeps(job):
	"""return all the job's dependencies"""
	res = defaultdict(set)
	for key, value in job.params.jobs.items():
		if value:
			value = expand_to_set(value)
			value = tuple(x for x in value if type(x) == Job)
			res['jobs.' + key].update(expand_to_set(value))
	for key, value in job.params.datasets.items():
		if value:
			value = expand_to_set(value)
			value = tuple(x for x in value if type(x) == Dataset)
			res['datasets.' + key].update(expand_to_set(value, lambda x: x.job))
	# options JobWithFile
	def recurse(options, name=''):
		if isinstance(options, JobWithFile):  # must happen before tuple
			res['jwf' + name].add(options.job)
		elif isinstance(options, dict):
			for key, val in options.items():
				recurse(val, '.'.join((name, key)))
		elif isinstance(options, (list, tuple, set)):
			for item in options:
				recurse(item, name)
	recurse(job.params.options)
	return res


def dsdeps(ds):
	"""return all the dataset's parents and previous"""
	res = defaultdict(set)
	if ds:
		if ds.parent:
			res['parent'].update(expand_to_set(ds.parent))
		if ds.previous:
			res['previous'].add(ds.previous)
	return res


class WrapperNode(unicode):
	def __init__(self, payload):
		self.payload = payload
		self.level = 0
		self.not_in_urdlist = False
		self.deps = set()
		self.subjobs = set()
		self.contextual_deps = set()
		self.depdict = {}
		self.done = False
		self.in_degrees = 0
		self.top_job = False
		self.neighbour_nodes = set()
		self.neighbour_edges = set()
		self.atmaxdepth = False  # Implemented in frontend but not in current backend.
		self.is_build = False

def try_subjobs(n):
	try:
		return n.payload.post.subjobs
	except FileNotFoundError:
		return {}

class Graph:
	def __init__(self, depsfun):
		self.count = 0
		self.nodes = {}
		self.edges = set()
		self.subjob_edges = set()
		self.depsfun = depsfun

	def get_or_create_node(self, item):
		"""get existing from item or create a new WrapperNode"""
		if item in self.nodes:
			return self.nodes[item]
		else:
			n = WrapperNode(item)
			n.safename = 'node' + str(self.count)
			n.depdict = self.depsfun(item)
			self.nodes[item] = n
			self.count += 1
			return n

	def create_edge(self, current, dep, relation=None):
		"""create a new edge (from WrapperNode to WrapperNode)"""
		if relation is not None:
			self.edges.add((current, dep, relation))
		else:
			self.subjob_edges.add((current, dep))

	def level2nodes(self):
		"""return {level: set_of_WrapperNodes_at_level}"""
		ret = defaultdict(set)
		for n in self.nodes.values():
			ret[n.level].add(n)
		return ret

	def populate_with_neighbours(self):
		"""add neighbour information to all WrapperNodes"""
		for src, dst, rel in self.edges:
			edgekey = src.safename + dst.safename
			src.neighbour_nodes.add(dst)
			dst.neighbour_nodes.add(src)
			src.neighbour_edges.add(edgekey)
			dst.neighbour_edges.add(edgekey)

	def convert_to_safe_strings(self):
		"""collapse neighbour_nodes and edges to strings, for later JSON output"""
		for n in self.nodes.values():
			n.neighbour_nodes = tuple(x.safename for x in sorted(n.neighbour_nodes))
		edges = defaultdict(set)
		for src, dst, rel in self.edges:
			edges[(src.safename, dst.safename)].add(rel)  # there can be more than one relation between the same two nodes
		self.edges = tuple((src, dst, ', '.join(sorted(v))) for (src, dst), v in edges.items())
		self.subjob_edges = tuple((src.safename, dst.safename) for (src, dst) in self.subjob_edges)
		self.nodes = {n.safename: n for n in self.nodes.values()}

	def depth_first(self, stack):
		"""
		Jobs and Datasets (but not JobLists) do a depth-first
		search to discover all nodes and number of inputs to each node
		"""
		while stack:
			current = stack.pop()
			if current.done:
				continue
			for relation, deps in current.depdict.items():
				for dep in deps:
					dep = self.get_or_create_node(dep)
					if dep != current:
						self.create_edge(current, dep, relation)
						current.deps.add(dep)
			for dep in sorted(current.deps):
				stack.append(dep)
				dep.in_degrees += 1
			if isinstance(current.payload, Job):
				current.subjobs = set(self.get_or_create_node(Job(jid)) for jid in try_subjobs(current).keys())
				for sjob in sorted(current.subjobs):
					self.create_edge(current, sjob)
					sjob.contextual_deps.add(current)
					stack.append(sjob)
					current.in_degrees += 1
			current.done = True

	def breadth_first(self, stack):
		"""breadth-first search to get the correct rendering level of each node."""
		while stack:
			current = stack.pop()
			if current.done:
				continue
			current.in_degrees -= 1
			if current.in_degrees <= 0:
				for dep in sorted(current.deps):
					dep.level = max(dep.level, current.level + 1)
					stack.append(dep)
				current.done = True


def create_graph(inputitem, urdinfo=()):
	if isinstance(inputitem, tuple):
		# Was JobList. Create and populate WrapperNodes for all jobs in input
		graph = Graph(jobdeps)
		inputitem = [graph.get_or_create_node(x) for x in inputitem]
		for n in inputitem:
			for relation, deps in n.depdict.items():
				for c in deps:
					c = graph.get_or_create_node(c)
					graph.create_edge(n, c, relation)
					c.in_degrees += 1
					n.deps.add(c)
		stack = [n for n in inputitem if n.in_degrees == 0]
	else:
		# Is Job or Dataset. Do depth-first to find in_degrees for all WrapperNodes.
		# in_degrees is then used later to compute each node's rendering level.
		graph = Graph(jobdeps if isinstance(inputitem, Job) else dsdeps)
		inputitem = graph.get_or_create_node(inputitem)
		inputitem.top_job = True
		stack = [inputitem, ]
		graph.depth_first(stack)
		# Reset the done-flag so it can be re-used in a second recursion.
		# Also add contextual_deps to deps.
		for n in graph.nodes.values():
			n.done = False
			for c in n.contextual_deps:
				if c not in n.deps:
					n.deps.add(c)
				else:
					c.in_degrees -= 1  # because it is already counted in the "original" deps
		reached_by_deps = set.union(*(n.deps for n in graph.nodes.values()))
		starters = set(graph.nodes.values()) - reached_by_deps
		stack = sorted(starters)
	graph.breadth_first(stack)

	# add parameters from payload (job/ds) to all WrapperNodes
	inputitem_set = set(inputitem)
	for n in graph.nodes.values():
		njob = n.payload if isinstance(n.payload, Job) else n.payload.job
		n.jobid = njob
		n.method = njob.method
		n.timestamp = datetime.fromtimestamp(njob.params.starttime).strftime("%Y-%m-%d %H:%M:%S")
		n.name = n.method
		if isinstance(n.payload, Job):
			n.files = sorted(n.payload.files())
			n.datasets = sorted(n.payload.datasets)
			n.subjobs = tuple((jid, Job(jid).method) for jid in try_subjobs(n))
			n.parent = njob.params.get('parent')
			n.is_build = n.payload.params.get('is_build', False)
			if urdinfo:
				jobid2urddep, jobid2name = urdinfo
				if n not in inputitem_set:
					n.not_in_urdlist = jobid2urddep.get(n.jobid, True)
				n.name = jobid2name.get(n.jobid, None)
		else:
			# dataset
			n.columns = sorted((colname, dscol.type) for colname, dscol in n.payload.columns.items())
			n.lines = "%d x % s" % (len(n.payload.columns), '{:,}'.format(sum(n.payload.lines)).replace(',', '_'))
			n.ds = n.payload
	graph.populate_with_neighbours()
	return graph


def placement(graph):
	class Ordering:
		"""The init function takes the first level of nodes as input.
		The update function takes each consecutive level of nodes as
		input.  It returns a list of the nodes in order.
		"""
		def __init__(self, nodes):
			self.order = {x: str(ix) for ix, x in enumerate(sorted(nodes))}
		def update(self, nodes):
			nodes = sorted(nodes, key=lambda x: self.order[x])
			for n in nodes:
				for ix, c in enumerate(sorted(n.deps)):
					if c not in self.order:
						self.order[c] = self.order[n] + str(ix)
				self.order.pop(n)
			for ix, (key, val) in enumerate(sorted(self.order.items(), key=lambda x: x[1])):
				self.order[key] = str(ix)
			return nodes
	# determine x and y coordinates
	level2nodes = graph.level2nodes()
	order = Ordering(level2nodes[0])
	for level, nodesatlevel in sorted(level2nodes.items()):
		nodesatlevel = order.update(nodesatlevel)
		for ix, n in enumerate(nodesatlevel):
			n.x = -160 * level + 32 * sin(ix)
			n.y =  140 * ix    + 50 * sin(level)
	# limit angles by adjusting x positions
	offset = {}
	for level, nodesatlevel in sorted(level2nodes.items()):
		maxangle = MAXANGLE
		xoffset = 0
		for n in nodesatlevel:
			for m in n.deps:
				dx = abs(n.x - m.x)
				dy = abs(n.y - m.y)
				angle = abs(atan(dy / dx))
				if angle > maxangle:
					maxangle = angle
					xoffset = (dy / tan(MAXANGLE)) - dx
		offset[level + 1] = xoffset
	totoffset = 0
	for level, xoffset in sorted(offset.items()):
		totoffset += xoffset
		for n in level2nodes[level]:
			n.x -= totoffset
	graph.convert_to_safe_strings()
	return dict(
		nodes=graph.nodes,
		edges=graph.edges,
		subjob_edges=graph.subjob_edges,
	)


def graph(inp, gtype):
	if gtype == 'urd':
		# jobid2urddep is {jobid: "urdlist/timestamp"} for all jobids in depending urdlist
		jobid2urddep = defaultdict(list)
		for key, urditem in inp.deps.items():
			for _, jid in urditem.joblist:
				jobid2urddep[jid].append("%s/%s" % (key, urditem.timestamp))
		jobid2urddep = {key: sorted(val) for key, val in jobid2urddep.items()}
		jlist = inp.joblist
		inp = tuple(Job(jid) for _, jid in jlist)
		jobid2name = {jid: name for name, jid in jlist}
		graph = create_graph(inp, urdinfo=(jobid2urddep, jobid2name))
	else:
		graph = create_graph(inp)
	ret = placement(graph)
	ret['type'] = 'job' if gtype in ('urd', 'job') else 'dataset'
	return ret



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2024 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test board adding metadata to files and decoding this with "ax sherlock".
'''

options = dict(
	command_prefix=['ax', '--config', '/some/path/here'],
)

from accelerator.compat import url_quote_more, urlopen, Request, HTTPError, unicode

from subprocess import Popen, check_output
import os
import struct
import time


def synthesis(job):
	os.putenv('XDG_CONFIG_HOME', job.path)   # make sure we can't be messed up by config
	os.putenv('ACCELERATOR_IGNORE_ETC', 'Y') # make sure we can't be messed up by global config
	socket_path = os.path.realpath('socket')
	p = Popen(options.command_prefix + ['board-server', socket_path])

	# Allow some time for board-server to start
	for _ in range(32):
		if os.path.exists(socket_path):
			break
		time.sleep(0.1)

	board_url = 'unixhttp://' + url_quote_more(socket_path) + '/'
	job_url = board_url + 'job/' + url_quote_more(job) + '/'

	# This is a cut down version of unixhttp.get() that doesn't do any
	# decoding of the response (as we will fetch binary files).
	def get(filename):
		# The path ends up in the Host-header if we don't override it, and
		# if the path contains characters that aren't in latin-1 that breaks.
		headers = {'Host': 'board'}
		url = job_url + filename
		req = Request(url, headers=headers)
		r = urlopen(req)
		try:
			resp = r.read()
			# It is inconsistent if we get HTTPError or not.
			# It seems we do when using TCP sockets, but not when using unix sockets.
			if r.getcode() >= 400:
				raise HTTPError(url, r.getcode(), resp, {}, None)
			return resp
		finally:
			try:
				r.close()
			except Exception:
				pass

	extra_metadata = br'{"job":"/\n/DoES/NoT.EXIST/NOPE","setup_hash":"nah","host":"cough","method":"none","time":0}'
	def mk_meta(prefix_a, prefix_b=b'', suffix=b'', offset=0):
		data = extra_metadata + suffix
		if isinstance(prefix_a, unicode):
			return struct.pack(prefix_a, len(data) + len(prefix_b) + offset) + prefix_b + data
		elif isinstance(prefix_b, unicode):
			return prefix_a + struct.pack(prefix_b, len(data) + len(prefix_a) + offset) + data
		else:
			return prefix_a + prefix_b + extra_metadata + suffix
	def crc(data):
		from zlib import crc32
		return struct.pack('>I', crc32(data) & 0xffffffff)

	# filename, contents, expected modified head, expected modified tail, has extra block
	# These need to be at least 20 bytes for board to try augmenting them.
	files = [
		('a.png', b'\x89PNG\x0d\x0a\x1a\x0a\x00\x00\x00\x00duMY\xae\x42\x60\x82\x00\x00\x00\x00IEND\xae\x42\x60\x82', 20, 8, False),
		('b.png', b'\x89PNG\x0d\x0a\x1a\x0a' + mk_meta('>I', b'tEXtExAx\0', offset=-4) + crc(b'tEXtExAx\0' + extra_metadata) + b'\x00\x00\x00\x00IEND\xae\x42\x60\x82', 80, 8, True),
		('a.jpg', b'\xff\xd8\xff\xea\x00\x10..............\xff\xff\xff\xd9', b'\xff\xd8\xff\xea\x00\x10..............\xff\xea', b'}\xff\xff\xff\xd9', False),
		('b.jpg', b'\xff\xd8\xff\xea' + mk_meta('>H', b'ExAx\0', offset=2) + b'\xff\xff\xff\xd9', 80, b'}\xff\xff\xff\xd9', True),
		('a.gif', b'GIF87ax................y;', b'GIF89ax................y', b'}\x00;', False),
		# Make sure the chunking decoder works.
		('b.gif', b'GIF89ax!\xff\x0bExAxMeta1.0\x20' + extra_metadata[:32] + struct.pack('<B', len(extra_metadata) - 32) + extra_metadata[32:] + b'\x00data;', 23 + len(extra_metadata) + 2, b'}\x00;', True),
		('a.avi', b'RIFF\x10\x00\x00\x00whatDUMY\x03\x00\x00\x00...\x00', 4, {b'}', b'}\0'}, False),
		('b.avi', b'RIFF' + struct.pack('<I', len(extra_metadata) + 12) + mk_meta(b'whatExAx', '<I', offset=-8), 4, {b'}', b'}\0'}, True),
		('a.mp4', b'\x00\x00\x00\x19ftypisom.............', b'\x00\x00', b'}', False),
		('b.mp4', b'\x00\x00\x00\x0cftypisom' + mk_meta('>I', b'ExAx', offset=4), b'\x00\x00', b'}', True),
		('a.mkv', b'\x1a\x45\xdf\xa3\xec\x40\x10.................', 24, b'}', False),
		('b.mkv', b'\x1a\x45\xdf\xa3\xec\x40' + mk_meta('<B', b'\xeaExAxMeta\x0e\x0a:'), 19 + len(extra_metadata), b'}', True),
		# PDF is stupid with newlines, so test a few versions. Also with and without (valid looking) startxref.
		('a.pdf', b'%PDF-1.5\nstartxref\n9\n%%EOF\n', b'%PDF-1.5\n% ExAx:{', b'}\nstartxref\n9\n%%EOF\n', False),
		('b.pdf', b'%PDF-1.5\r\nnotxref\r9\r%%EOF', b'%PDF-1.5\r\nnotxref\r9\n% ExAx:{', b'}\n%%EOF', False),
		('c.pdf', b'%PDF-1.0\n% ExAx:' + extra_metadata + b'\nnotxref\n9\rstartxref\r\n9\r\n%%EOF', 26 + len(extra_metadata), b'}\nstartxref\r\n9\r\n%%EOF', True),
		('d.pdf', b'%PDF-1.777\r\n% ExAx:' + extra_metadata + b'\nstartxref\r\n9ad\r%%EOF\r\n', 34 + len(extra_metadata), b'}\n%%EOF\r\n', True), # bad startxref - data after
		('e.pdf', b'%PDF-3.9\n' + b'.' * 8000 + b'!\rstartxref\n17\n%%EOF\n', b'%PDF-3.9\n' + b'.' * 8000 + b'!\n% ExAx:{', b'}\nstartxref\n17\n%%EOF\n', False), # >4k
		('a.jxl', b'\xff\x0a...naked JPEG XL...', b'\0\0\0\x0cJXL \r\n\x87\n\0\0\0\x14ftypjxl \0\0\0\0jxl \0\0\0\x1djxlc\xff\x0a...naked JPEG XL...', b'}', False),
		('b.jxl', b'\0\0\0\x0cJXL \r\n\x87\n\0\0\0\x14ftypjxl \0\0\0\0jxl \0\0\0\x0ajxlc\xff\x0a', 42, b'}', False),
		('c.jxl', b'\0\0\0\x0cJXL \r\n\x87\n\0\0\0\x14ftypjxl \0\0\0\0jxl \0\0\0\x0ajxlc\xff\x0a' + mk_meta('>I', b'ExAx', offset=4), 50 + len(extra_metadata), b'}', True),
	]

	for filename, contents, want_head, want_tail, has_extra_block in files:
		if isinstance(want_head, int):
			want_head = contents[:want_head]
		if isinstance(want_tail, int):
			want_tail = contents[-want_tail:]
		with open(filename, 'wb') as fh:
			fh.write(contents)
		modified_contents = get(filename)
		if contents == modified_contents:
			raise Exception('%s was not modified by board' % (filename,))
		filename = 'modified.' + filename
		with open(filename, 'wb') as fh:
			fh.write(modified_contents)
		if not any(modified_contents.startswith(v) for v in (want_head if isinstance(want_head, set) else (want_head,))):
			raise Exception('Expected %s to start with %r, but it did not' % (filename, want_head,))
		if not any(modified_contents.endswith(v) for v in (want_tail if isinstance(want_tail, set) else (want_tail,))):
			raise Exception('Expected %s to end with %r, but it did not' % (filename, want_tail,))
		got = check_output(options.command_prefix + ['sherlock', filename])
		got = got.decode('utf-8').strip()
		if has_extra_block:
			want_jobs = 'NOPE unknown job, does not exist on disk\n' + job
		else:
			want_jobs = job
		if got != want_jobs:
			raise Exception('Expected "ax sherlock %s" to give %r, got %r' % (filename, want_jobs, got,))

	p.terminate()
	p.wait()



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2024 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

# Functions for generating, inserting and extracting metadata about the job
# that built files.

from __future__ import print_function

import json
import re
import socket
import struct
import sys
import traceback
import zlib

from accelerator.compat import PY3, FileNotFoundError
from accelerator.extras import json_decode

if PY3:
	crc32 = zlib.crc32
else:
	def crc32(data):
		return zlib.crc32(data) & 0xffffffff


def b64hash_setup(filename):
	from base64 import b64encode
	from hashlib import sha1
	try:
		with open(filename, 'rb') as fh:
			data = fh.read()
	except FileNotFoundError:
		return None
	# Try to remove the part that differs between finished and unfinished jobs.
	data = data.split(b'\n\n', 1)[-1]
	return b64encode(sha1(data).digest(), b'_-').rstrip(b'=').decode('ascii')

def job_metadata(job):
	d = {
		'job': job.path,
		'method': job.method,
		'time': int(job.params.starttime), # who needs sub-second precision?
		'setup_hash': b64hash_setup(job.filename('setup.json')),
		'host': socket.gethostname(),
	}
	res = json.dumps(d)
	if PY3:
		res = res.encode('ascii')
	return res


def generate_png(fh, job, size):
	payload = b'tEXtExAx\0' + job_metadata(job)
	crc = crc32(payload)
	data = (
		struct.pack('>I', len(payload) - 4) +
		payload +
		struct.pack('>I', crc) +
		b'\x00\x00\x00\x00IEND\xae\x42\x60\x82'
	)
	return [
		(0, size - 12),
		(data, len(data)),
	]

def extract_png(fh):
	fh.seek(0)
	if fh.read(8) != b'\x89PNG\x0d\x0a\x1a\x0a':
		return
	while True:
		data = fh.read(4)
		if len(data) != 4:
			return
		z, = struct.unpack('>I', data)
		chunk_id = fh.read(4)
		if chunk_id == b'tEXt':
			data = fh.read(z)
			crc, = struct.unpack('>I', fh.read(4))
			if crc == crc32(b'tEXt' + data) and data.startswith(b'ExAx\0'):
				yield data[5:]
		else:
			fh.seek(z + 4, 1)


def generate_jpeg(fh, job, size):
	# It's probably a JFIF or similar, so skip all initial APP blocks.
	pos = 2
	fh.seek(pos)
	while True:
		data = fh.read(4)
		if len(data) != 4:
			return
		typ, z = struct.unpack('>HH', data)
		if typ < 0xff00 or z < 2:
			return
		if typ < 0xffe0 or typ > 0xffef:
			break
		pos += z + 2
		fh.seek(pos, 0)
	payload = b'ExAx\0' + job_metadata(job)
	data = struct.pack('>HH', 0xffea, len(payload) + 2) + payload
	return [
		(0, pos),
		(data, len(data)),
		(pos, size - pos),
	]

def extract_jpeg(fh):
	fh.seek(0)
	data = fh.read(2)
	if data != b'\xff\xd8':
		return
	pos = 2
	# Look at APP blocks until we run out.
	while True:
		data = fh.read(4)
		if len(data) != 4:
			return
		typ, z = struct.unpack('>HH', data)
		if typ < 0xffe0 or typ > 0xffef or z < 2:
			# Not an APP block, so we stop looking.
			return
		pos += z + 2
		if typ == 0xffea:
			data = fh.read(z - 2)
			if len(data) == z - 2 and data.startswith(b'ExAx\0'):
				yield data[5:]
		fh.seek(pos, 0)


def generate_gif(fh, job, size):
	def chunks(data):
		while data:
			part = data[:255]
			data = data[255:]
			yield struct.pack('<B', len(part)) + part
	payload = job_metadata(job)
	# The first chunk of the extension block has to be exactly 11 bytes, so
	# says the specification. 8 bytes Identifier + 3 bytes Authentication Code.
	data = b'!\xff\x0bExAxMeta1.0' + b''.join(chunks(payload)) + b'\x00;'
	return [
		(b'GIF89a', 6), # might have been GIF87a, which does not support application extension blocks.
		(6, size - 7),
		(data, len(data)),
	]

def extract_gif(fh):
	def unchunk(pos):
		while pos + 1 < len(data):
			z, = struct.unpack('<B', data[pos:pos + 1]) # "z = pos[0]" but python2 compatible.
			if z == 0:
				return
			yield data[pos + 1:pos + z + 1]
			pos += z + 1
	fh.seek(0, 2)
	fh.seek(max(0, fh.tell() - 8192), 0)
	data = fh.read()
	signature = b'!\xff\x0bExAxMeta1.0'
	pos = data.find(signature)
	while pos >= 0 and pos + 16 < len(data):
		yield b''.join(unchunk(pos + 14))
		pos = data.find(signature, pos + 17)


def detect_riff(data, size):
	if data.startswith(b'RIFF') and len(data) >= 8:
		return struct.unpack('<I', data[4:8])[0] + 8 == size

def generate_riff(fh, job, size):
	fh.seek(0)
	data = fh.read(4)
	if data != b'RIFF':
		return
	data = fh.read(4)
	if len(data) != 4:
		return
	z, = struct.unpack('<I', data)
	if z + 8 != size:
		# Incomplete file / garbage after.
		return
	payload = job_metadata(job)
	data = b'ExAx' + struct.pack('<I', len(payload)) + payload
	if len(data) % 2 == 1:
		data += b'\0'
	if z + len(data) >= 0xfffffff8:
		# File would become too big.
		return
	return [
		(b'RIFF' + struct.pack('<I', z + len(data)), 8),
		(8, z),
		(data, len(data)),
	]

def extract_riff(fh):
	fh.seek(0)
	data = fh.read(4)
	if data != b'RIFF':
		return
	data = fh.read(4)
	if len(data) != 4:
		return
	z, = struct.unpack('<I', data)
	fh.seek(0, 2)
	if z + 8 != fh.tell():
		# incomplete file / garbage after
		return
	fh.seek(12)
	while True:
		fourcc = fh.read(4)
		data = fh.read(4)
		if len(data) != 4:
			return
		z, = struct.unpack('<I', data)
		if fourcc == b'ExAx':
			data = fh.read(z)
			if len(data) == z:
				yield data
		else:
			fh.seek(z, 1)
		if z % 2 == 1:
			fh.seek(1, 1)


def generate_isom(fh, job, size, pos=0):
	fh.seek(pos)
	data = fh.read(8)
	if len(data) != 8:
		return
	z, fourcc = struct.unpack('>II', data)
	if fourcc != 0x66747970: # "ftyp"
		return
	pos += z
	while True:
		if pos == size:
			# Found the end of the file without problems, let's put our data here.
			payload = job_metadata(job)
			data = struct.pack('>I', len(payload) + 8) + b'ExAx' + payload
			return [
				(0, size),
				(data, len(data)),
			]
		fh.seek(pos)
		data = fh.read(8)
		if len(data) != 8:
			return
		z, fourcc = struct.unpack('>II', data)
		pos += z

def extract_isom(fh, pos=0):
	fh.seek(pos)
	data = fh.read(8)
	if len(data) != 8:
		return
	z, fourcc = struct.unpack('>II', data)
	if fourcc != 0x66747970: # "ftyp"
		return
	pos += z
	while True:
		fh.seek(pos)
		data = fh.read(8)
		if len(data) != 8:
			return
		z, fourcc = struct.unpack('>II', data)
		if fourcc == 0x45784178: # "ExAx"
			data = fh.read(z - 8)
			if len(data) == z - 8:
				yield data
		pos += z


# Simplify parsing by having enough of a header to just look for that instead of proper EBML parsing.
ebml_signature = b'\xeaExAxMeta\x0e\x0a:'

def generate_ebml(fh, job, size):
	payload = ebml_signature + job_metadata(job)
	if len(payload) > 0x3ffe:
		return
	# This looks like a valid EBML Void element, but as we are just
	# placing it after the file this probably does not matter.
	data = struct.pack('>BH', 0xec, len(payload) | 0x4000) + payload
	return [
		(0, size),
		(data, len(data)),
	]

def extract_ebml(fh):
	fh.seek(0, 2)
	fh.seek(max(0, fh.tell() - 8192), 0)
	data = fh.read()
	pos = data.find(ebml_signature)
	while pos >= 2:
		z = data[pos - 2:pos]
		z, = struct.unpack('>H', z)
		if z & 0xc000 != 0x4000:
			pos = data.find(ebml_signature, pos + 12)
			continue
		z &= 0x3fff
		if z > 0x3ffe or pos + z > len(data):
			return
		pos += len(ebml_signature)
		z -= len(ebml_signature)
		yield data[pos:pos + z]
		pos = data.find(ebml_signature, pos + z)


def generate_pdf(fh, job, size):
	offset = max(0, size - 4096)
	fh.seek(offset, 0)
	filetail = fh.read()
	m = re.search(br'[\r\n]+startxref(?:(?:[ \t\f\v]*%.*)?[\r\n]+)+\d+(?:(?:[ \t\f\v]*%.*)?[\r\n]+)+%%EOF[\r\n]*$', filetail)
	if not m:
		# There should be a startxref just before %%EOF, but if there
		# isn't we'll place our comment directly before the %%EOF marker.
		m = re.search(br'[\r\n]+%%EOF[\r\n]*$', filetail)
	if m:
		pos0 = pos1 = m.start()
		if not pos0: # in case we started in the middle of a newline
			return
		while filetail[pos1] in b'\r\n':
			pos1 += 1
		data = b'\n% ExAx:' + job_metadata(job) + b'\n'
		pos0 += offset
		pos1 += offset
		return [
			(0, pos0),
			(data, len(data)),
			(pos1, size - pos1),
		]

def extract_pdf(fh):
	fh.seek(0, 2)
	fh.seek(max(0, fh.tell() - 8192), 0)
	data = fh.read()
	signature = b'\n% ExAx:{'
	pos = data.find(signature)
	while pos >= 0:
		endpos = data.find(b'\n', pos + len(signature))
		yield data[pos + len(signature) - 1:endpos]
		pos = data.find(signature, pos + 1)


def generate_jxl_naked(fh, job, size):
	fh.seek(0)
	data = fh.read(2)
	if data != b'\xff\x0a':
		return
	# Naked JPEG XL does not allow extra blocks, so we have to box it.
	header = b'\0\0\0\x0cJXL \r\n\x87\n\0\0\0\x14ftypjxl \0\0\0\0jxl ' + \
		struct.pack('>I', size + 8) + b'jxlc'
	payload = job_metadata(job)
	data = struct.pack('>I', len(payload) + 8) + b'ExAx' + payload
	return [
		(header, len(header)),
		(0, size),
		(data, len(data)),
	]

def generate_jxl_boxed(fh, job, size):
	fh.seek(0)
	data = fh.read(12)
	if data != b'\0\0\0\x0cJXL \r\n\x87\n':
		return
	return generate_isom(fh, job, size, 12)

def extract_jxl(fh):
	fh.seek(0)
	data = fh.read(12)
	if data != b'\0\0\0\x0cJXL \r\n\x87\n':
		return
	return extract_isom(fh, 12)


formats = [
	(
		'PNG', None, generate_png, extract_png,
		b'\x89PNG\x0d\x0a\x1a\x0a',
		b'\x00\x00\x00\x00IEND\xae\x42\x60\x82',
	),
	(
		'JPEG', None, generate_jpeg, extract_jpeg,
		b'\xff\xd8\xff',
		b'\xff\xd9',
	),
	(
		'GIF', '.gif', generate_gif, extract_gif,
		re.compile(br'GIF8[79]a'),
		b';',
	),
	(
		'RIFF based format (AVI, WEBP, ...)', None, generate_riff, extract_riff,
		detect_riff,
		b'',
	),
	(
		'ISO media format (MP4, HEIF, ...)', None, generate_isom, extract_isom,
		re.compile(br'\0...ftyp', re.DOTALL),
		b'',
	),
	(
		'EBML based format (MKV, WEBM, ...)', None, generate_ebml, extract_ebml,
		b'\x1a\x45\xdf\xa3',
		b'',
	),
	(
		'PDF', '.pdf', generate_pdf, extract_pdf,
		re.compile(br'%PDF-\d\.\d'),
		re.compile(br'.*[\r\n]%%EOF[\r\n]*$', re.DOTALL),
	),
	(
		'JPEG XL (naked codestream)', '.jxl', generate_jxl_naked, None,
		b'\xff\x0a',
		b'',
	),
	(
		'JPEG XL (boxed)', None, generate_jxl_boxed, extract_jxl,
		re.compile(b'\0\0\0\x0cJXL \r\n\x87\n\0...ftyp', re.DOTALL),
		b'',
	),
]

def matcher(pattern, where):
	if hasattr(pattern, 'match'):
		return lambda v, _: pattern.match(v)
	elif callable(pattern):
		return pattern
	elif where == 'start':
		return lambda v, _: v.startswith(pattern)
	elif where == 'end':
		return lambda v, _: v.endswith(pattern)

formats = [
	(name, ext, generate, extract, matcher(header, 'start'), matcher(trailer, 'end'))
	for name, ext, generate, extract, header, trailer in formats
]

def insert_metadata(filename, fh, job, size):
	res = None
	if size > 20 and job:
		fh.seek(0, 0)
		start = fh.read(20)
		fh.seek(-20, 2)
		end = fh.read(20)
		for name, ext, generate, extract, header, trailer in formats:
			if ext and not filename.lower().endswith(ext):
				continue
			if header(start, size) and trailer(end, size):
				try:
					res = generate(fh, job, size)
				except Exception:
					traceback.print_exc(file=sys.stderr)
				if res:
					break
				print("Failed to generate %s metadata." % (name,), file=sys.stderr)
	return res or [(0, size)]

def extract_metadata(filename, fh):
	def decode(gen):
		for data in gen:
			try:
				yield json_decode(data.decode('utf-8'))
			except Exception:
				yield None
	fh.seek(0, 0)
	start = fh.read(20)
	if len(start) != 20:
		return
	fh.seek(-20, 2)
	size = fh.tell() + 20
	end = fh.read(20)
	for name, ext, generate, extract, header, trailer in formats:
		if not extract:
			continue
		if ext and not filename.lower().endswith(ext):
			continue
		if header(start, size) and trailer(end, size):
			res = extract(fh)
			if res:
				res = decode(res)
			return res



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2021-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import gzip
from random import randint
import struct
from accelerator._dsutil import siphash24
from accelerator.dsutil import typed_writer
from accelerator.subjobs import build

description = r'''
Test the number column type, verifying that both DatasetWriter and
dataset_type produce the expected bitstream.
'''

def synthesis(job, slices):
	# All the representations we want to verify.
	values = [
		# 1 byte values
		[i, '=B', i + 128 + 5]
		for i in range(-5, 123)
	] + [
		# 3 bytes values
		[-6, '=bh', 2, -6],
		[123, '=bh', 2, 123],
		[-0x8000, '=bh', 2, -0x8000],
		[0x7fff, '=bh', 2, 0x7fff],
		# 5 byte values
		[-0x8001, '=bi', 4, -0x8001],
		[0x8000, '=bi', 4, 0x8000],
		[-0x80000000, '=bi', 4, -0x80000000],
		[0x7fffffff, '=bi', 4, 0x7fffffff],
		# 9 byte values
		[-0x80000001, '=bq', 8, -0x80000001],
		[0x80000000, '=bq', 8, 0x80000000],
		[-0x8000000000000000, '=bq', 8, -0x8000000000000000],
		[0x7fffffffffffffff, '=bq', 8, 0x7fffffffffffffff],
		# special values
		[None, '=b', 0],
		[0.1, '=bd', 1, 0.1],
		[9.2, '=bd', 1, 9.2],
		[3.3, '=bd', 1, 3.3],
		[1.4, '=bd', 1, 1.4],
		[-0.5, '=bd', 1, -0.5],
	]

	# We also want to verify the encoding of larger values, so here is a
	# python implementation of the expected encoding and a few values to try.
	def encode_as_big_number(num):
		len_bytes = num.bit_length() // 8 + 1;
		digits = []
		for _ in range(len_bytes):
			digits.append(num & 0xff)
			num >>= 8
		assert 8 < len(digits) < 127
		digits.insert(0, len(digits))
		return b''.join(struct.pack('=B', d) for d in digits)

	values.extend([
		# Smallest values that use the big encoding.
		[0x8000000000000000],
		[-0x8000000000000001],
	])
	for e in range(64, 1007):
		num = 2 ** e
		for offset in (-2, -1, 0, 1, 2):
			values.append([offset + num])
			values.append([offset - num])
		values.append([randint(3, num // 2) + num])
		values.append([randint(3, num // 2) - num])
	# And finally the biggest possible values.
	values.extend([[(2 ** 1007) - 1], [-(2 ** 1007) + 1]])

	# Verify each value through a manual typed_writer.
	# Also write to a dataset, a csv and a value2bytes dict.
	value2bytes = {}
	dw = job.datasetwriter()
	dw.add('num', 'number', none_support=True)
	write = dw.get_split_write()
	with job.open('data.csv', 'wt') as csv_fh:
		csv_fh.write('num\n')
		for v in values:
			value = v[0]
			write(value)
			csv_fh.write('%s\n' % (value,))
			if len(v) == 1:
				want_bytes = encode_as_big_number(v[0])
			else:
				want_bytes = struct.pack(*v[1:])
			value2bytes[value] = want_bytes
			with typed_writer('number')('tmp', compression='gzip', none_support=True) as w:
				w.write(value)
			with gzip.open('tmp', 'rb') as fh:
				got_bytes = fh.read()
			assert want_bytes == got_bytes, "%r gave %r, wanted %r" % (value, got_bytes, want_bytes,)

	# Make sure we get the same representation through a dataset.
	# Assumes that the column is merged (a single file for all slices).
	ds = dw.finish()
	just_values = set(v[0] for v in values)
	assert set(ds.iterate(None, 'num')) == just_values, "Dataset contains wrong values"
	want_bytes = b''.join(value2bytes[v] for v in ds.iterate(None, 'num'))
	with gzip.open(ds.column_filename('num'), 'rb') as fh:
		got_bytes = fh.read()
	assert want_bytes == got_bytes, "All individual encoding are right, but not in a dataset?"

	# csvimport and dataset_type the same thing,
	# verify we got the same bytes
	csvimport = build('csvimport', filename=job.filename('data.csv'))
	jid = build('dataset_type', source=csvimport, column2type={'num': 'number'}, defaults={'num': None})
	ds_typed = jid.dataset()
	with gzip.open(ds_typed.column_filename('num'), 'rb') as fh:
		got_bytes = fh.read()
	assert want_bytes == got_bytes, "csvimport + dataset_type (%s) gave different bytes" % (jid,)

	# Also test the hashing is as expected, both in DatasetWriter, dataset_type and dataset_hashpart.
	def hash_num(num):
		if not num:
			return 0
		if isinstance(num, float):
			enc = struct.pack('=d', num)
		elif -0x8000000000000000 <= num < 0x8000000000000000:
			enc = struct.pack('=q', num)
		else:
			enc = encode_as_big_number(num)[1:] # excluding the length
		return siphash24(enc)

	# Test some actual values to be the expected ones.
	# Don't hard code anything that doesn't use the big encoding, as not
	# only the hash value but the hashed data is endian dependant there.
	dw_hash = typed_writer('number').hash
	def endianfix(v):
		return struct.unpack('<Q', struct.pack('=Q', v))[0]
	assert endianfix(dw_hash(0xBEEFC0DEBEEFC0DEBEEFC0DE1337)) == 0x938a366370bb5bb5
	# But the special cased 0 values are ok to test.
	assert dw_hash(0) == 0
	assert dw_hash(0.0) == 0
	assert dw_hash(None) == 0
	# And we can check that number and the int and float types agree.
	assert dw_hash(42) == typed_writer('int32').hash(42)
	assert dw_hash(42) == typed_writer('float32').hash(42.0) # int and intable float should agree.
	assert dw_hash(4.2) == typed_writer('float64').hash(4.2)
	assert dw_hash(0xBEEFC0DEB16) == typed_writer('int64').hash(0xBEEFC0DEB16)

	dw = job.datasetwriter(name='hashed', hashlabel='num')
	dw.add('num', 'number', none_support=True)
	write = dw.get_split_write()
	per_slice = [set() for _ in range(slices)]
	for v in just_values:
		write(v)
		per_slice[dw_hash(v) % slices].add(v)
		assert hash_num(v) == dw_hash(v), v
	ds_local = dw.finish()
	ds_hashed = build('dataset_hashpart', source=ds_typed, hashlabel='num').dataset()
	ds_typehashed = build('dataset_type', source=csvimport, column2type={'num': 'number'}, defaults={'num': None}, hashlabel='num').dataset()
	for sliceno in range(slices):
		for hashname, ds in [
			('DatasetWriter', ds_local),
			('dataset_hashpart', ds_hashed),
			('dataset_type', ds_typehashed),
		]:
			assert set(ds.iterate(sliceno, 'num')) == per_slice[sliceno], 'wrong in slice %d in %s (hashed by %s)' % (sliceno, ds, hashname)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2024 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import division
from __future__ import absolute_import

description = r'''
Rewrite a dataset (or chain to previous) with new hashlabel.
'''

from shutil import copyfileobj
from os.path import exists

from accelerator import OptionString

options = {
	'hashlabel'                 : OptionString,
	'caption'                   : '"%(caption)s" hashed on %(hashlabel)s',
	'length'                    : -1, # Go back at most this many datasets. You almost always want -1 (which goes until previous.source)
	'chain_slices'              : False, # one dataset per slice (avoids rewriting at the end)
}

datasets = ('source', 'previous',)

def prepare(job, slices):
	previous = datasets.previous
	res = []
	chain = datasets.source.chain(stop_ds={datasets.previous: 'source'}, length=options.length)
	if not chain:
		raise Exception("previous had the same source - this job makes no sense")
	for ix, ds in enumerate(chain):
		res.append(prepare_one(ix, ds, previous, job, chain, slices))
		previous = res[-1][0][-1] # dws[-1]
	return chain, res

def prepare_one(ix, source, previous, job, chain, slices):
	caption = options.caption % dict(caption=source.caption, hashlabel=options.hashlabel)
	filename = source.filename
	dws = []
	for sliceno in range(slices):
		if sliceno == slices - 1 and options.chain_slices and ix == len(chain) - 1:
			name = "default"
		else:
			name = '%d.%d' % (ix, sliceno,)
		dw = job.datasetwriter(
			caption="%s (slice %d)" % (caption, sliceno),
			hashlabel=options.hashlabel,
			filename=filename,
			previous=previous,
			name=name,
			for_single_slice=sliceno,
			copy_mode=True,
		)
		previous = dw
		dws.append(dw)
	names = []
	cols = {}
	for n, c in source.columns.items():
		# names has to be in the same order as the add calls
		# so the iterator returns the same order the writer expects.
		names.append(n)
		cols[n] = (c.type, chain.none_support(n))
		for dw in dws:
			if dw:
				dw.add(n, c.type, none_support=cols[n][1])
	return dws, names, caption, filename, cols

def analysis(sliceno, prepare_res):
	chain, p_res = prepare_res
	for ds, p in zip(chain, p_res):
		analysis_one(sliceno, ds, p)

def analysis_one(sliceno, source, prepare_res):
	dws, names = prepare_res[:2]
	write = dws[sliceno].get_split_write_list()
	for values in source.iterate(sliceno, names, copy_mode=True):
		write(values)

def synthesis(prepare_res, job, slices):
	if not options.chain_slices:
		chain, p_res = prepare_res
		previous = datasets.previous
		for ix, p in enumerate(p_res):
			previous = synthesis_one(ix, chain, p, job, slices, previous)

def synthesis_one(ix, chain, prepare_res, job, slices, previous):
	# If we don't want a chain we abuse our knowledge of dataset internals
	# to avoid recompressing. Don't do this stuff yourself.
	dws, names, caption, filename, cols = prepare_res
	merged_dw = job.datasetwriter(
		name='default' if ix == len(chain) - 1 else str(ix),
		caption=caption,
		hashlabel=options.hashlabel,
		filename=filename,
		previous=previous,
		meta_only=True,
		columns=cols,
	)
	merged_dw.set_compressions(dws[0]._compressions)
	for sliceno in range(slices):
		merged_dw.set_lines(sliceno, sum(dw._lens[sliceno] for dw in dws))
		for dwno, dw in enumerate(dws):
			merged_dw.set_minmax((sliceno, dwno), dw._minmax[sliceno])
		for n in names:
			fn = merged_dw.column_filename(n, sliceno=sliceno)
			with open(fn, "wb") as out_fh:
				for dw in dws:
					fn = dw.column_filename(n, sliceno=sliceno)
					if exists(fn):
						with open(fn, "rb") as in_fh:
							copyfileobj(in_fh, out_fh)
	for dw in dws:
		dw.discard()
	return merged_dw



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2022-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test all the ax sub-commands at least a little bit.
'''

options = dict(
	command_prefix=['ax', '--config', '/some/path/here'],
)

import os
from subprocess import check_output

# this is for job --source to have a second file
depend_extra = ('test_shell_commands.txt',)

def ax(cmd):
	cmd = options.command_prefix + cmd
	print(cmd)
	res = check_output(cmd)
	res = res.decode('utf-8', 'replace')
	return res

all_checked = set()

def chk(cmd, want_in_help=[], want_in_call=[], dont_want_in_call=[]):
	all_checked.add(cmd[0])
	for args, want_l, dont_want_l in (
		([cmd[0], '-h'], ['--help'] + want_in_help, []),
		(cmd, want_in_call, dont_want_in_call),
	):
		if not want_l and not dont_want_l:
			continue
		output = ax(args)
		for want in want_l:
			if want not in output:
				print("Expected to find %r in %r output:\n\n%s\n\nbut didn't." % (want, args, output,))
				raise Exception("Failed in command %r" % (cmd[0],))
		for dont_want in dont_want_l:
			if dont_want in output:
				print("Did not expect to find %r in %r output:\n\n%s\n\nbut did." % (dont_want, args, output,))
				raise Exception("Failed in command %r" % (cmd[0],))

def synthesis(job):
	print('look for this later')
	with job.open('filename', 'w') as fh:
		in_written_file = 'stuff\nin\na\nfile\n' # \n is not literal here - not seen in this file
		fh.write(in_written_file)
	dw = job.datasetwriter(columns={'this is a column': 'ascii'})
	dw.get_split_write()('this is a value')
	ds = dw.finish()
	os.putenv('XDG_CONFIG_HOME', job.path) # make sure we can't be messed up by config
	os.putenv('ACCELERATOR_IGNORE_ETC', 'Y') # make sure we can't be messed up by global config
	os.unsetenv('CLICOLOR_FORCE')
	chk(['-h'], want_in_help=['\ncommands:\n', '\naliases:\n'])
	chk(['abort'], want_in_help=['--quiet'])
	chk(['alias', 'cat'], want_in_help=['shows all aliases'], want_in_call=["grep -e ''"])
	chk(['board-server'], want_in_help=['localhost:8520'])
	chk(['ds', '--', ds], want_in_help=['--list'], want_in_call=['Columns:', 'this is a column'])
	chk(['gc'], want_in_help=['--dry-run'])
	chk(['grep', '--', '', ds], want_in_help=['--tab-length'], want_in_call=['this is a value'])
	chk(['hist', '--', ds, 'this is a column'], want_in_help=['show at most this many'], want_in_call=['this is a value\t1'])
	chk(['init', '--no-git', job.filename('projdir')], want_in_help=['--no-git'], want_in_call=[''])
	assert os.path.isdir('projdir/workdirs')
	chk(['intro'], want_in_call=['ax init --examples', 'ax script'])
	chk(['job', '--', job], want_in_help=['--just-output', ':urdlist:[entry]'], want_in_call=['"command_prefix":'])
	chk(['job', '--just-output', '--', job], want_in_call=['look for this later'], dont_want_in_call=['most other things'])
	in_this_file = 'this string is only here, and will be found here'
	in_extra = 'Magic cookie: ' + 'Delicious!' # different formatting here - not seen in this file
	extra_hdr = 'test_shell_commands.txt\n==============' # \n is not literal here - not seen in this file
	chk(['job', '--source', '--', job], want_in_call=[in_this_file, in_extra, extra_hdr]),
	chk(['job', '--source-file=*.txt', '--', job], want_in_call=[in_extra], dont_want_in_call=[extra_hdr, in_this_file]),
	chk(['job', '--file=filename', '--', job], want_in_call=[in_written_file], dont_want_in_call=['===', in_this_file]),
	chk(['method', 'csvimport'], want_in_help=['lists methods'], want_in_call=['filename'])
	chk(['run'], want_in_help=['[script ...]', 'WORKDIR'])
	chk(['script', 'build_tests'], want_in_help=['describes build scripts'], want_in_call=['Needs at least 3 slices to work.'])
	chk(['server'], want_in_help=['--debuggable'])
	chk(['sherlock'], want_in_help=['downloaded through board'])
	chk(['status'], want_in_help=['--short'], want_in_call=['test_shell_commands (', 'synthesis (', 'TAIL OF OUTPUT:'])
	chk(['status', '--short'], want_in_call=['test_shell_commands synthesis ('])
	chk(['urd'], want_in_help=[':urdlist:[entry]'], want_in_call=['/tests.urd'])
	chk(['urd', 'tests.urd/'], want_in_call=['2021-09-27T03:14'])
	chk(['urd', 'tests.urd/2021-09-27T03:14'], want_in_call=['test_shell_data'])
	chk(['urd-server'], want_in_help=['--allow-passwordless'])
	chk(['version'], want_in_call=['Running on '])
	chk(['workdir'], want_in_help=['--full-path'], want_in_call=[job.workdir])
	chk(['workdir', '--', job.workdir], want_in_call=[job, 'test_shell_commands'])

	ax_help = ax(['--help'])
	cmd_list = ax_help.split('\ncommands:\n', 1)[1].split('\naliases:\n', 1)[0]
	cmd_list = {cmd.split()[0] for cmd in cmd_list.strip().split('\n')}
	missed = cmd_list - all_checked
	if missed:
		raise Exception("Didn't check the following commands: " + ' '.join(sorted(missed)))
	not_in_help = all_checked - cmd_list - {'-h', '--help'}
	if not_in_help:
		raise Exception("Missing in help output: " + ' '.join(sorted(not_in_help)))



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2022-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test that the JobError exception contains the correct status stack
and line number from dataset iteration.
'''

options = dict(
	die_in=str,
	die_on_lines=[int],
	status1=str,
	status2=str,
)

jobs = ('source',)


from accelerator import subjobs, status, JobError
import re

def iterate(line_a):
	stop_v = line_a[0]
	line_a = line_a[1:]
	for v in jobs.source.dataset('dont_see_this').iterate(0, 'a'):
		pass
	for v in jobs.source.dataset(str(len(line_a))).iterate(0, 'a'):
		if v == stop_v:
			if line_a:
				return iterate(line_a)
			else:
				if options.status2:
					with status(options.status2):
						1/0
				else:
					1/0

def die_here():
	if options.die_on_lines:
		if options.status1:
			with status(options.status1):
				iterate(options.die_on_lines)
		else:
			iterate(options.die_on_lines)
	else:
		with status(options.status1):
			1/0

def mk_ds(job, name):
	dw = job.datasetwriter(name=name, columns={'a': 'int32'}, allow_missing_slices=True)
	dw.set_slice(0)
	for v in range(1, 1000):
		dw.write(v)
	dw.finish()

def prepare(job):
	if not options.die_in:
		mk_ds(job, '0')
		mk_ds(job, '1')
		mk_ds(job, '2')
		mk_ds(job, 'dont_see_this')
	elif options.die_in == 'prepare':
		die_here()

def analysis(sliceno, job):
	if options.die_in == 'analysis' and sliceno == 1:
		die_here()

def synthesis(job):
	if not options.die_in:
		for where, on_line, status1, status2 in (
			('prepare', [], 'status without datasets', None),
			('prepare', [1], None, None),
			('analysis', [643, 84], 'status with two datasets', None),
			('synthesis', [999, 57, 350], None, 'status inside three datasets'),
			('analysis', [811], None, None),
			('prepare', [99, 999], 'before', 'after')
		):
			try:
				jid = subjobs.build(
					'test_status_in_exceptions',
					die_in=where,
					die_on_lines=on_line,
					status1=status1,
					status2=status2,
					source=job,
				)
				raise Exception('Should not have built job ' + jid)
			except JobError as e:
				got = e.format_msg()
				assert 'dont_see_this' not in got, got
				want_re = 'Status when the exception occurred:.*' + where
				if status1:
					want_re += '.*' + status1
				for ix, want_line in enumerate(on_line, 1):
					ds = job.dataset(str(len(on_line) - ix))
					want_re += '.*Iterating %s:0 reached line %d\n' % (re.escape(ds.quoted), want_line,)
				if status2:
					want_re += '.*' + status2
				assert re.search(want_re, got, re.DOTALL), got
				if not on_line:
					assert 'Iterating ' not in got, got
	elif options.die_in == 'synthesis':
		die_here()
